from math import pi

import numpy as np
from ase.units import Bohr, Ha
from gpaw.core import PWDesc, UGDesc
from gpaw.core.plane_waves import PWArray
from gpaw.new import cached_property
from gpaw.new.poisson import PoissonSolver
from scipy.special import erf


def make_poisson_solver(pw: PWDesc,
                        grid: UGDesc,
                        pbc_c,
                        charge: float,
                        strength: float = 1.0,
                        dipolelayer: bool = False,
                        **kwargs) -> PoissonSolver:
    if charge != 0.0 and not pbc_c.any():
        return ChargedPWPoissonSolver(pw, grid, charge, strength, **kwargs)

    ps = PWPoissonSolver(pw, charge, strength)

    if dipolelayer:
        return DipoleLayerPWPoissonSolver(ps, grid, **kwargs)
    assert not kwargs
    return ps


class PWPoissonSolver(PoissonSolver):
    def __init__(self,
                 pw: PWDesc,
                 charge: float = 0.0,
                 strength: float = 1.0):
        self.pw = pw
        self.charge = charge
        self.strength = strength

        self.ekin_g = pw.ekin_G.copy()
        if pw.comm.rank == 0:
            # Avoid division by zero:
            self.ekin_g[0] = 1.0

    def __str__(self) -> str:
        txt = ('poisson solver:\n'
               f'  ecut: {self.pw.ecut * Ha}  # eV\n')
        if self.strength != 1.0:
            txt += f'  strength: {self.strength}\n'
        if self.charge != 0.0:
            txt += f'  uniform background charge: {self.charge}  # electrons\n'
        return txt

    def solve(self,
              vHt_g: PWArray,
              rhot_g: PWArray) -> float:
        """Solve Poisson equeation.

        Places result in vHt_g ndarray.
        """
        epot = self._solve(vHt_g, rhot_g)
        return epot

    def _solve(self,
               vHt_g,
               rhot_g) -> float:
        vHt_g.data[:] = 2 * pi * self.strength * rhot_g.data
        if self.pw.comm.rank == 0:
            # Use uniform backgroud charge in case we have a charged system:
            vHt_g.data[0] = 0.0
        if not isinstance(self.ekin_g, vHt_g.xp.ndarray):
            self.ekin_g = vHt_g.xp.array(self.ekin_g)
        vHt_g.data /= self.ekin_g
        epot = 0.5 * vHt_g.integrate(rhot_g)
        return epot


class ChargedPWPoissonSolver(PWPoissonSolver):
    def __init__(self,
                 pw: PWDesc,
                 grid: UGDesc,
                 charge: float,
                 strength: float = 1.0,
                 alpha: float = None,
                 eps: float = 1e-5):
        """Reciprocal-space Poisson solver for charged molecules.

        * Add a compensating Guassian-shaped charge to the density
          in order to make the total charge neutral (placed in the
          middle of the unit cell

        * Solve Poisson equation.

        * Correct potential so that it has the correct 1/r
          asymptotic behavior

        * Correct energy to remove the artificial interaction with
          the compensation charge

        Parameters
        ----------
        pw: PWDesc
        grid: UGDesc
        charge: float
        strength: float
        alpha: float
        eps: float

        Attributes
        ----------
        alpha : float
        charge_g : np.ndarray
            Guassian-shaped charge in reciprocal space
        potential_g : PWArray
             Potential in reciprocal space created by charge_g
        """
        super().__init__(pw, charge, strength)

        if alpha is None:
            # Shortest distance from center to edge of cell:
            rcut = 0.5 / (pw.icell**2).sum(axis=1).max()**0.5

            # Make sure e^(-alpha*rcut^2)=eps:
            alpha = -rcut**-2 * np.log(eps)

        self.alpha = alpha

        center_v = pw.cell_cv.sum(axis=0) / 2
        G2_g = 2 * pw.ekin_G
        G_gv = pw.G_plus_k_Gv
        self.charge_g = np.exp(-1 / (4 * alpha) * G2_g +
                               1j * (G_gv @ center_v))
        self.charge_g *= charge / pw.dv

        R_Rv = grid.xyz()
        d_R = ((R_Rv - center_v)**2).sum(axis=3)**0.5
        potential_R = grid.empty()

        # avoid division by 0
        zero_indx = d_R == 0
        d_R[zero_indx] = 1
        potential_R.data[:] = charge * erf(alpha**0.5 * d_R) / d_R
        # at zero we should have:
        # erf(alpha**0.5 * d_R) / d_R = alpha**0.5 * 2 / sqrt(pi)
        potential_R.data[zero_indx] = charge * alpha**0.5 * 2 / np.sqrt(pi)
        self.potential_g = potential_R.fft(pw=pw)

    def __str__(self) -> str:
        txt, x, _ = super().__str__().rsplit('\n', 2)
        assert x.startswith('  uniform background charge:')
        txt += (
            '\n  # using Gaussian-shaped compensation charge: e^(-alpha r^2)\n'
            f'  alpha: {self.alpha}   # bohr^-2')
        return txt

    def _solve(self,
               vHt_g,
               rhot_g) -> float:
        neutral_g = rhot_g.copy()
        neutral_g.data += self.charge_g

        if neutral_g.desc.comm.rank == 0:
            error = neutral_g.data[0]  # * self.pd.gd.dv
            assert error.imag == 0.0, error
            assert abs(error.real) < 0.00001, error
            neutral_g.data[0] = 0.0

        vHt_g.data[:] = 2 * pi * neutral_g.data
        vHt_g.data /= self.ekin_g
        epot = 0.5 * vHt_g.integrate(neutral_g)
        epot -= self.potential_g.integrate(rhot_g)
        epot -= self.charge**2 * (self.alpha / 2 / pi)**0.5
        vHt_g.data -= self.potential_g.data
        return epot


class DipoleLayerPWPoissonSolver(PoissonSolver):
    def __init__(self,
                 ps: PWPoissonSolver,
                 grid: UGDesc,
                 width: float = 1.0):  # Ångström
        self.ps = ps
        self.grid = grid
        self.width = width / Bohr
        (self.axis,) = np.where(~grid.pbc_c)

    def solve(self,
              vHt_g: PWArray,
              rhot_g: PWArray) -> float:
        epot = self.ps.solve(vHt_g, rhot_g)

        dip_v = -rhot_g.moment()
        c = self.axis
        L = self.grid.cell_cv[c, c]
        self.correction = 2 * np.pi * dip_v[c] * L / self.grid.volume
        vHt_g.data -= 2 * self.correction * self.sawtooth_g.data

        return epot + 2 * np.pi * dip_v[c]**2 / self.grid.volume

    @cached_property
    def sawtooth_g(self) -> PWArray:
        grid = self.grid
        if grid.comm.rank == 0:
            c = self.axis
            L = grid.cell_cv[c, c]
            w = self.width / 2
            assert w < L / 2
            gc = int(w / grid.size_c[c])
            x = np.linspace(0, L, grid.size_c[c], endpoint=False)
            sawtooth = x / L - 0.5
            a = 1 / L - 0.75 / w
            b = 0.25 / w**3
            sawtooth[:gc] = x[:gc] * (a + b * x[:gc]**2)
            sawtooth[-gc:] = -sawtooth[gc:0:-1]
            sawtooth_r = grid.new(comm=None).empty()
            shape = [1, 1, 1]
            shape[c] = -1
            sawtooth_r.data[:] = sawtooth.reshape(shape)
            sawtooth_g = sawtooth_r.fft(pw=self.ps.pw).data
        else:
            sawtooth_g = None

        result_g = self.ps.pw.empty()
        result_g.scatter_from(sawtooth_g)
        return result_g



from __future__ import annotations

import numpy as np

from ase.units import Ha

from gpaw.pw.descriptor import PWMapping

from gpaw.response.pw_parallelization import (Blocks1D,
                                              PlaneWaveBlockDistributor)
from gpaw.response.frequencies import (FrequencyDescriptor,
                                       ComplexFrequencyDescriptor)
from gpaw.response.pair_functions import (SingleQPWDescriptor,
                                          map_ZgG_array_to_reduced_pd)


class Chi0RelatedData:
    """Base class for chi0 related data objects.

    Right now, all we do is to limit boiler plate code..."""

    def __init__(self,
                 wd: FrequencyDescriptor,
                 qpd: SingleQPWDescriptor):
        self.wd = wd
        self.qpd = qpd
        self.q_c = qpd.q_c

        # Basis set size
        self.nG = qpd.ngmax
        self.nw = len(wd)


class Chi0BodyData(Chi0RelatedData):
    """Data object containing the response body data arrays
    for a single q-point, while holding also the corresponding
    basis descriptors and block distributor."""

    def __init__(self, wd, qpd,
                 blockdist: PlaneWaveBlockDistributor):
        super().__init__(wd, qpd)

        # Initialize block distibution of plane wave basis
        self.blockdist = blockdist
        self.blocks1d = Blocks1D(blockdist.blockcomm, self.nG)

        # Data array
        self.data_WgG = self.zeros()

    def zeros(self):
        return np.zeros(self.WgG_shape, complex)

    @property
    def mynG(self):
        return self.blocks1d.nlocal
    
    @property
    def WgG_shape(self):
        return (self.nw, self.mynG, self.nG)

    def get_distributed_frequencies_array(self):
        """Copy data to a 'wGG'-like array, distributed over the entire world.

        This differs from copy_array_with_distribution('wGG'), in that the
        frequencies are distributed over world, instead of among the block
        communicator."""
        return self.blockdist.distribute_frequencies(self.data_WgG, self.nw)

    def copy_array_with_distribution(self, distribution):
        """Copy data to a new array of a desired distribution.

        Parameters
        ----------
        distribution: str
            Array distribution. Choices: 'wGG' and 'WgG'
        """
        data_x = self.blockdist.distribute_as(self.data_WgG, self.nw,
                                              distribution)

        if data_x is self.data_WgG:
            # When asking for 'WgG' distribution or when there is no block
            # distribution at all, we may still be pointing to the original
            # array, but we want strictly to return a copy
            assert distribution == 'WgG' or \
                self.blockdist.blockcomm.size == 1
            data_x = self.data_WgG.copy()

        return data_x

    def copy_with_reduced_pd(self, qpd):
        """Make a copy corresponding to a new plane-wave description."""
        new_chi0_body = Chi0BodyData(self.wd, qpd, self.blockdist)

        # Map data to reduced plane-wave representation
        new_chi0_body.data_WgG[:] = map_ZgG_array_to_reduced_pd(
            self.qpd, qpd, self.blockdist, self.data_WgG)

        return new_chi0_body


class Chi0DrudeData:
    def __init__(self, zd: ComplexFrequencyDescriptor):
        self.zd = zd
        self.plasmafreq_vv, self.chi_Zvv = self.zeros()

    def zeros(self):
        return (np.zeros(self.vv_shape, complex),  # plasmafreq
                np.zeros(self.Zvv_shape, complex))  # chi0_drude

    @staticmethod
    def from_frequency_descriptor(wd, rate):
        """Construct the Chi0DrudeData object from a frequency descriptor and
        the imaginary part (in eV) of the resulting horizontal frequency
        contour"""
        rate = rate / Ha  # eV -> Hartree
        zd = ComplexFrequencyDescriptor(wd.omega_w + 1.j * rate)

        return Chi0DrudeData(zd)

    @property
    def nz(self):
        return len(self.zd)

    @property
    def vv_shape(self):
        return (3, 3)

    @property
    def Zvv_shape(self):
        return (self.nz,) + self.vv_shape


class Chi0OpticalExtensionData(Chi0RelatedData):
    def __init__(self, wd, qpd):
        assert qpd.optical_limit
        super().__init__(wd, qpd)

        self.head_Wvv, self.wings_WxvG = self.zeros()
        
    def zeros(self):
        return (np.zeros(self.Wvv_shape, complex),  # head
                np.zeros(self.WxvG_shape, complex))  # wings

    @property
    def Wvv_shape(self):
        return (self.nw, 3, 3)

    @property
    def WxvG_shape(self):
        return (self.nw, 2, 3, self.nG)

    def copy_with_reduced_pd(self, qpd):
        """Make a copy corresponding to a new plane-wave description."""
        new_chi0_optical_extension = Chi0OpticalExtensionData(self.wd, qpd)

        # Copy the head (present in any plane-wave representation)
        new_chi0_optical_extension.head_Wvv[:] = self.head_Wvv

        # Map the wings to the reduced plane-wave description
        G2_G1 = PWMapping(qpd, self.qpd).G2_G1
        new_chi0_optical_extension.wings_WxvG[:] \
            = self.wings_WxvG[..., G2_G1]

        return new_chi0_optical_extension


class Chi0Data(Chi0RelatedData):
    """Container object for the chi0 data objects for a single q-point,
    while holding also the corresponding basis descriptors and block
    distributor."""

    def __init__(self,
                 chi0_body: Chi0BodyData,
                 chi0_opt_ext: Chi0OpticalExtensionData | None = None):
        super().__init__(chi0_body.wd, chi0_body.qpd)
        self.body = chi0_body

        self.optical_limit = self.qpd.optical_limit
        if self.optical_limit:
            assert isinstance(chi0_opt_ext, Chi0OpticalExtensionData)
            assert chi0_opt_ext.wd is self.wd
            assert chi0_opt_ext.qpd is self.qpd
        else:
            assert chi0_opt_ext is None
        self.optical_extension = chi0_opt_ext

    @staticmethod
    def from_chi0_body_data(chi0_body):
        """Construct the container from a chi0 body data instance."""
        qpd = chi0_body.qpd
        if qpd.optical_limit:
            wd = chi0_body.wd
            chi0_optical_extension = Chi0OpticalExtensionData(wd, qpd)
        else:
            chi0_optical_extension = None

        return Chi0Data(chi0_body, chi0_optical_extension)

    def copy_with_reduced_pd(self, qpd):
        """Make a copy of the data object, reducing the plane wave basis."""
        new_body = self.body.copy_with_reduced_pd(qpd)
        if self.optical_limit:
            new_optical_extension = \
                self.optical_extension.copy_with_reduced_pd(qpd)
        else:
            new_optical_extension = None

        return Chi0Data(new_body, new_optical_extension)

    @property
    def chi0_WgG(self):
        return self.body.data_WgG

    @property
    def chi0_Wvv(self):
        if self.optical_limit:
            return self.optical_extension.head_Wvv

    @property
    def chi0_WxvG(self):
        if self.optical_limit:
            return self.optical_extension.wings_WxvG



import pytest
from ase import Atoms
from ase.units import Hartree

from gpaw import GPAW


@pytest.mark.later
def test_fileio_wfs_io(in_tmp_dir):
    h2 = Atoms('H2', [(0, 0, 0), (0, 0, 1)])
    h2.center(vacuum=2.0)
    calc = GPAW(mode='fd', nbands=2, convergence={'eigenstates': 1e-3})
    h2.calc = calc
    h2.get_potential_energy()
    r0 = calc.wfs.eigensolver.error * Hartree**2 / 2
    assert r0 < 1e-3
    calc.write('h2', 'all')

    # refine the restart file containing the wfs
    calc = GPAW('h2', convergence={'eigenstates': 1e-5})
    calc.get_atoms().get_potential_energy()
    r1 = calc.wfs.eigensolver.error * Hartree**2 / 2
    assert r1 < 1e-5



"""
A class for finding optimal
orbitals of the KS-DFT or PZ-SIC
functionals using exponential transformation
direct minimization

arXiv:2101.12597 [physics.comp-ph]
Comput. Phys. Commun. 267, 108047 (2021).
https://doi.org/10.1016/j.cpc.2021.108047
"""


import numpy as np
import warnings
from gpaw.directmin.tools import expm_ed, expm_ed_unit_inv
from gpaw.directmin.lcao.directmin_lcao import DirectMinLCAO
from scipy.linalg import expm
from gpaw.directmin import search_direction, line_search_algorithm
from gpaw.directmin.functional import get_functional
from gpaw import BadParallelization
from copy import deepcopy


class ETDM:

    """
    Exponential Transformation Direct Minimization (ETDM)
    """

    def __init__(self,
                 searchdir_algo='l-bfgs-p',
                 linesearch_algo='swc-awc',
                 partial_diagonalizer='Davidson',
                 update_ref_orbs_counter=20,
                 update_ref_orbs_canonical=False,
                 update_precond_counter=1000,
                 use_prec=True, matrix_exp='pade-approx',
                 representation='sparse',
                 functional='ks',
                 orthonormalization='gramschmidt',
                 randomizeorbitals=False,
                 checkgraderror=False,
                 localizationtype=None,
                 need_localization=True,
                 need_init_orbs=True,
                 constraints=None,
                 ):
        """
        This class performs the exponential transformation
        direct minimization:
        E = E[C_ref e^{A}]
        C_ref are the reference orbitals
        A is a skew-Hermitian matrix
        We want to find A that optimizes the orbitals. The optimal orbitals
        are those corresponding to a minimum for the ground state or a saddle
        point for excited states.

        :param searchdir_algo: algorithm for calculating the search direction
        (e.g.LBFGS)
        :param linesearch_algo: line search (e.g. strong Wolfe conditions)
        :param partial_diagonalizer: Algorithm to use for partial
        diagonalization of the electronic Hessian if DO-GMF is used
        :param update_ref_orbs_counter: When to update C_ref
        :param update_ref_orbs_canonical: update C_ref to canonical orbitals
        :param update_precond_counter: when to update the preconditioner
        :param use_prec: use preconditioner or not
        :param matrix_exp: algorithm for calculating the matrix exponential and
        gradient. Can be one of 'pade-approx', 'egdecomp', 'egdecomp-u-invar'
        (used with u-invar representation)
        :param representation: the way the elements of A are stored. Can be one
        of 'sparse', 'full', 'u-invar'
        :param functional: KS or PZ-SIC
        :param orthonormalization: orthonormalize the orbitals using
        Gram-Schmidt or Loewdin orthogonalization, or getting the orbitals as
        eigenstates of the Hamiltonian matrix. Can be one of 'gramschmidt',
        'loewdin', 'diag'
        :param randomizeorbitals: if True, add noise to the initial guess
        :param checkgraderror: check error in estimation of gradient
        :param localizationtype: Foster-Boys, Pipek-Mezey, Edm.-Rudenb.
        :param need_localization: use localized orbitals as initial guess
        :param need_init_orbs: if false, then use orbitals stored in kpt
        :param constraints: List of constraints for each kpt. Can be given
        either as pairs of orbital indices or single indices which are
        converted to list of all pairs involving the single index
        """

        assert representation in ['sparse', 'u-invar', 'full'], 'Value Error'
        assert matrix_exp in ['egdecomp', 'egdecomp-u-invar', 'pade-approx'], \
            'Value Error'
        if matrix_exp == 'egdecomp-u-invar':
            assert representation == 'u-invar', 'Use u-invar representation ' \
                                                'with egdecomp-u-invar'
        assert orthonormalization in ['gramschmidt', 'loewdin', 'diag'], \
            'Value Error'

        self.localizationtype = localizationtype
        self.eg_count = 0
        self.update_ref_orbs_counter = update_ref_orbs_counter
        self.update_ref_orbs_canonical = update_ref_orbs_canonical
        self.update_precond_counter = update_precond_counter
        self.use_prec = use_prec
        self.matrix_exp = matrix_exp
        self.iters = 0
        self.restart = False
        self.name = 'etdm'
        self.localizationtype = localizationtype
        self.need_localization = need_localization
        self.need_init_orbs = need_init_orbs
        self.randomizeorbitals = randomizeorbitals
        self.representation = representation
        self.orthonormalization = orthonormalization
        self.constraints = constraints

        self.gmf = False
        self.searchdir_algo = search_direction(
            searchdir_algo, self, partial_diagonalizer)
        sd_name = self.searchdir_algo.name.split('_')
        if sd_name[0] == 'l-bfgs-p' and not self.use_prec:
            raise ValueError('Use l-bfgs-p with use_prec=True')
        if len(sd_name) == 2:
            if sd_name[1] == 'gmf':
                self.searchdir_algo.name = sd_name[0]
                self.gmf = True
                self.g_vec_u_original = None
                self.pd = partial_diagonalizer

        self.line_search = line_search_algorithm(linesearch_algo,
                                                 self.evaluate_phi_and_der_phi,
                                                 self.searchdir_algo)
        self.func = get_functional(functional)

        self.checkgraderror = checkgraderror
        self._norm_commutator, self._norm_grad = 0., 0.
        self.error = 0

        # these are things we cannot initialize now
        self.dtype = None
        self.nkpts = None
        self.gd = None
        self.nbands = None

        # values: vectors of the elements of matrices, keys: kpt number
        self.a_vec_u = {}  # for the elements of the skew-Hermitian matrix A
        self.g_vec_u = {}  # for the elements of the gradient matrix G
        self.evecs = {}   # eigenvectors for i*a_vec_u
        self.evals = {}   # eigenvalues for i*a_vec_u
        self.ind_up = {}
        self.n_dim = {}
        self.alpha = 1.0  # step length
        self.phi_2i = [None, None]  # energy at last two iterations
        self.der_phi_2i = [None, None]  # energy gradient w.r.t. alpha
        self.hess = {}  # approximate Hessian

        # for mom
        self.initial_occupation_numbers = None

        # in this attribute we store the object specific to each mode
        self.dm_helper = None

        self.initialized = False

    def __repr__(self):

        sda_name = self.searchdir_algo.name
        lsa_name = self.line_search.name

        add = ''
        pd_add = ''
        if self.gmf:
            add = ' with minimum mode following'
            pardi = {'Davidson': 'Finite difference generalized Davidson '
                     'algorithm'}
            pd_add = '       ' \
                     'Partial diagonalizer: {}\n'.format(
                         pardi[self.pd['name']])

        sds = {'sd': 'Steepest Descent',
               'fr-cg': 'Fletcher-Reeves conj. grad. method',
               'quick-min': 'Molecular-dynamics based algorithm',
               'l-bfgs': 'L-BFGS algorithm',
               'l-bfgs-p': 'L-BFGS algorithm with preconditioning',
               'l-sr1p': 'Limited-memory SR1P algorithm'}

        lss = {'max-step': 'step size equals one',
               'parabola': 'Parabolic line search',
               'swc-awc': 'Inexact line search based on cubic interpolation,\n'
                          '                    strong and approximate Wolfe '
                          'conditions'}

        repr_string = 'Direct minimisation' + add + ' using exponential ' \
                      'transformation.\n'
        repr_string += '       ' \
                       'Search ' \
                       'direction: {}\n'.format(sds[sda_name] + add)
        repr_string += pd_add
        repr_string += '       ' \
                       'Line ' \
                       'search: {}\n'.format(lss[lsa_name])
        repr_string += '       ' \
                       'Preconditioning: {}\n'.format(self.use_prec)
        repr_string += '       ' \
                       'WARNING: do not use it for metals as ' \
                       'occupation numbers are\n' \
                       '                ' \
                       'not found variationally\n'

        return repr_string

    def initialize(self, gd, dtype, nbands, nkpts, nao, using_blacs,
                   bd_comm_size, kpt_u):

        assert nbands == nao, \
            'Please, use: nbands=\'nao\''
        if not bd_comm_size == 1:
            raise BadParallelization(
                'Band parallelization is not supported')
        if using_blacs:
            raise BadParallelization(
                'ScaLapack parallelization is not supported')

        self.gd = gd
        self.dtype = dtype
        self.nbands = nbands
        self.nkpts = nkpts

        for kpt in kpt_u:
            u = self.kpointval(kpt)
            # dimensionality of the problem.
            # this implementation rotates among all bands
            self.n_dim[u] = self.nbands

        self.initialized = True

        # need reference orbitals
        self.dm_helper = None

    def initialize_dm_helper(self, wfs, ham, dens):

        self.dm_helper = DirectMinLCAO(
            wfs, ham, self.nkpts,
            diagonalizer=None,
            orthonormalization=self.orthonormalization,
            need_init_orbs=self.need_init_orbs
        )
        self.need_init_orbs = self.dm_helper.need_init_orbs
        # mom
        wfs.calculate_occupation_numbers(dens.fixed)
        occ_name = getattr(wfs.occupations, "name", None)
        if occ_name == 'mom':
            self.initial_occupation_numbers = wfs.occupations.numbers.copy()
            self.initialize_mom(wfs, dens)

        for kpt in wfs.kpt_u:
            f_unique = np.unique(kpt.f_n)
            if len(f_unique) > 2 and self.representation == 'u-invar':
                warnings.warn("Use representation == 'sparse' when "
                              "there are unequally occupied orbitals "
                              "as the functional is not unitary invariant")

        # randomize orbitals?
        if self.randomizeorbitals:
            for kpt in wfs.kpt_u:
                self.randomize_orbitals_kpt(wfs, kpt)
            self.randomizeorbitals = False

        # initialize matrices
        self.set_variable_matrices(wfs.kpt_u)
        # if no empty state no need to optimize
        for k in self.ind_up:
            if not self.ind_up[k][0].size or not self.ind_up[k][1].size:
                self.n_dim[k] = 0
        # set reference orbitals
        self.dm_helper.set_reference_orbitals(wfs, self.n_dim)

    def set_variable_matrices(self, kpt_u):

        # Matrices are sparse and Skew-Hermitian.
        # They have this structure:
        #  A_BigMatrix =
        #
        # (  A_1          A_2 )
        # ( -A_2.T.conj() 0   )
        #
        # where 0 is a zero-matrix of size of (M-N) * (M-N)
        #
        # A_1 i skew-hermitian matrix of N * N,
        # N-number of occupied states
        # A_2 is matrix of size of (M-N) * N,
        # M - number of basis functions
        #
        # if the energy functional is unitary invariant
        # then A_1 = 0
        # (see Hutter J et. al, J. Chem. Phys. 101, 3862 (1994))
        #
        # We will keep A_1 as we would like to work with metals,
        # SIC, and molecules with different occupation numbers.
        # this corresponds to 'sparse' representation
        #
        # Thus, for the 'sparse' we need to store upper
        # triangular part of A_1, and matrix A_2, so in total
        # (M-N) * N + N * (N - 1)/2 = N * (M - (N + 1)/2) elements
        #
        # we will store these elements as a vector and
        # also will store indices of the A_BigMatrix
        # which correspond to these elements.
        #
        # 'u-invar' corresponds to the case when we want to
        # store only A_2, that is this representation is sparser

        for kpt in kpt_u:
            n_occ = get_n_occ(kpt)
            u = self.kpointval(kpt)
            # M - one dimension of the A_BigMatrix
            M = self.n_dim[u]
            if self.representation == 'u-invar':
                i1, i2 = [], []
                for i in range(n_occ):
                    for j in range(n_occ, M):
                        i1.append(i)
                        i2.append(j)
                self.ind_up[u] = (np.asarray(i1), np.asarray(i2))
            else:
                if self.representation == 'full' and self.dtype == complex:
                    # Take indices of all upper triangular and diagonal
                    # elements of A_BigMatrix
                    self.ind_up[u] = np.triu_indices(self.n_dim[u])
                else:
                    self.ind_up[u] = np.triu_indices(self.n_dim[u], 1)
                    if self.representation == 'sparse':
                        # Delete indices of elements that correspond
                        # to 0 matrix in A_BigMatrix
                        zero_ind = -((M - n_occ) * (M - n_occ - 1)) // 2
                        if zero_ind == 0:
                            zero_ind = None
                        self.ind_up[u] = (self.ind_up[u][0][:zero_ind].copy(),
                                          self.ind_up[u][1][:zero_ind].copy())

            shape_of_arr = len(self.ind_up[u][0])

            self.a_vec_u[u] = np.zeros(shape=shape_of_arr, dtype=self.dtype)
            self.g_vec_u[u] = np.zeros(shape=shape_of_arr, dtype=self.dtype)
            self.evecs[u] = None
            self.evals[u] = None

            # All constraints passed as a list of a single orbital index are
            # converted to all lists of orbital pairs involving that orbital
            # index
            if self.constraints:
                self.constraints[u] = convert_constraints(
                    self.constraints[u], self.n_dim[u],
                    len(kpt.f_n[kpt.f_n > 1e-10]), self.representation)

        # This conversion makes it so that constraint-related functions can
        # iterate through a list of no constraints rather than checking for
        # None every time
        if self.constraints is None:
            self.constraints = [[] for _ in range(len(kpt_u))]

        self.iters = 1

    def iterate(self, ham, wfs, dens):
        """
        One iteration of direct optimization
        for occupied orbitals

        :param ham:
        :param wfs:
        :param dens:
        :return:
        """
        with wfs.timer('Direct Minimisation step'):
            self.update_ref_orbitals(wfs, ham, dens)

            a_vec_u = self.a_vec_u
            n_dim = self.n_dim
            alpha = self.alpha
            phi_2i = self.phi_2i
            der_phi_2i = self.der_phi_2i
            c_ref = self.dm_helper.reference_orbitals

            if self.iters == 1:
                phi_2i[0], g_vec_u = \
                    self.get_energy_and_gradients(a_vec_u, n_dim, ham, wfs,
                                                  dens, c_ref)
            else:
                g_vec_u = self.g_vec_u_original if self.gmf else self.g_vec_u

            make_pd = False
            if self.gmf:
                with wfs.timer('Partial Hessian diagonalization'):
                    self.searchdir_algo.update_eigenpairs(
                        g_vec_u, wfs, ham, dens)
                # The diagonal Hessian approximation must be positive-definite
                make_pd = True

            with wfs.timer('Preconditioning:'):
                precond = self.get_preconditioning(
                    wfs, self.use_prec, make_pd=make_pd)

            with wfs.timer('Get Search Direction'):
                # calculate search direction according to chosen
                # optimization algorithm (e.g. L-BFGS)
                p_vec_u = self.searchdir_algo.update_data(wfs, a_vec_u,
                                                          g_vec_u, precond)

            # recalculate derivative with new search direction
            der_phi_2i[0] = 0.0
            for k in g_vec_u:
                der_phi_2i[0] += g_vec_u[k].conj() @ p_vec_u[k]
            der_phi_2i[0] = der_phi_2i[0].real
            der_phi_2i[0] = wfs.kd.comm.sum(der_phi_2i[0])

            alpha, phi_alpha, der_phi_alpha, g_vec_u = \
                self.line_search.step_length_update(a_vec_u, p_vec_u,
                                                    n_dim, ham, wfs, dens,
                                                    c_ref,
                                                    phi_0=phi_2i[0],
                                                    der_phi_0=der_phi_2i[0],
                                                    phi_old=phi_2i[1],
                                                    der_phi_old=der_phi_2i[1],
                                                    alpha_max=5.0,
                                                    alpha_old=alpha,
                                                    kpdescr=wfs.kd)

            if wfs.gd.comm.size > 1:
                with wfs.timer('Broadcast gradients'):
                    alpha_phi_der_phi = np.array([alpha, phi_2i[0],
                                                  der_phi_2i[0]])
                    wfs.gd.comm.broadcast(alpha_phi_der_phi, 0)
                    alpha = alpha_phi_der_phi[0]
                    phi_2i[0] = alpha_phi_der_phi[1]
                    der_phi_2i[0] = alpha_phi_der_phi[2]
                    for kpt in wfs.kpt_u:
                        k = self.kpointval(kpt)
                        wfs.gd.comm.broadcast(g_vec_u[k], 0)

            # calculate new matrices for optimal step length
            for k in a_vec_u:
                a_vec_u[k] += alpha * p_vec_u[k]
            self.alpha = alpha
            self.g_vec_u = g_vec_u
            self.iters += 1

            # and 'shift' phi, der_phi for the next iteration
            phi_2i[1], der_phi_2i[1] = phi_2i[0], der_phi_2i[0]
            phi_2i[0], der_phi_2i[0] = phi_alpha, der_phi_alpha,

    def get_energy_and_gradients(self, a_vec_u, n_dim, ham, wfs, dens,
                                 c_ref):

        """
        Energy E = E[C_ref exp(A)]. Gradients G_ij[C, A] = dE/dA_ij

        :param wfs:
        :param ham:
        :param dens:
        :param a_vec_u: A
        :param c_ref: C_ref
        :param n_dim:
        :return:
        """

        self.rotate_wavefunctions(wfs, a_vec_u, n_dim, c_ref)

        e_total = self.update_ks_energy(ham, wfs, dens)

        with wfs.timer('Calculate gradients'):
            g_vec_u = {}
            self.error = 0.0
            self.e_sic = 0.0  # this is odd energy
            for kpt in wfs.kpt_u:
                k = self.kpointval(kpt)
                if n_dim[k] == 0:
                    g_vec_u[k] = np.zeros_like(a_vec_u[k])
                    continue
                g_vec_u[k], error = self.dm_helper.calc_grad(
                    wfs, ham, kpt, self.func, self.evecs[k], self.evals[k],
                    self.matrix_exp, self.representation, self.ind_up[k],
                    self.constraints[k])

                self.error += error
            self.error = wfs.kd.comm.sum(self.error)
            self.e_sic = wfs.kd.comm.sum(self.e_sic)

        self.eg_count += 1

        if self.representation == 'full' and self.checkgraderror:
            self._norm_commutator = 0.0
            for kpt in wfs.kpt_u:
                u = self.kpointval(kpt)
                a_mat = vec2skewmat(a_vec_u[u], self.n_dim[u],
                                    self.ind_up[u], wfs.dtype)
                g_mat = vec2skewmat(g_vec_u[u], self.n_dim[u],
                                    self.ind_up[u], wfs.dtype)

                tmp = np.linalg.norm(g_mat @ a_mat - a_mat @ g_mat)
                if self._norm_commutator < tmp:
                    self._norm_commutator = tmp

                tmp = np.linalg.norm(g_mat)
                if self._norm_grad < tmp:
                    self._norm_grad = tmp

        return e_total + self.e_sic, g_vec_u

    def update_ks_energy(self, ham, wfs, dens):
        """
        Update Kohn-Sham energy
        It assumes the temperature is zero K.
        """

        dens.update(wfs)
        ham.update(dens, wfs, False)

        return ham.get_energy(0.0, wfs, False)

    def evaluate_phi_and_der_phi(self, a_vec_u, p_mat_u, n_dim, alpha,
                                 ham, wfs, dens, c_ref,
                                 phi=None, g_vec_u=None):
        """
        phi = f(x_k + alpha_k*p_k)
        der_phi = \\grad f(x_k + alpha_k*p_k) \\cdot p_k
        :return:  phi, der_phi # floats
        """
        if phi is None or g_vec_u is None:
            x_mat_u = {k: a_vec_u[k] + alpha * p_mat_u[k] for k in a_vec_u}
            phi, g_vec_u = self.get_energy_and_gradients(x_mat_u, n_dim,
                                                         ham, wfs, dens, c_ref)

            # If GMF is used save the original gradient and invert the parallel
            # projection onto the eigenvectors with negative eigenvalues
            if self.gmf:
                self.g_vec_u_original = deepcopy(g_vec_u)
                g_vec_u = self.searchdir_algo.invert_parallel_grad(g_vec_u)

        der_phi = 0.0
        for k in p_mat_u:
            der_phi += g_vec_u[k].conj() @ p_mat_u[k]

        der_phi = der_phi.real
        der_phi = wfs.kd.comm.sum(der_phi)

        return phi, der_phi, g_vec_u

    def update_ref_orbitals(self, wfs, ham, dens):
        """
        Update reference orbitals

        :param wfs:
        :param ham:
        :return:
        """

        if self.representation == 'full':
            badgrad = self._norm_commutator > self._norm_grad / 3. and \
                self.checkgraderror
        else:
            badgrad = False
        counter = self.update_ref_orbs_counter
        if (self.iters % counter == 0 and self.iters > 1) or \
                (self.restart and self.iters > 1) or badgrad:
            self.iters = 1
            if self.update_ref_orbs_canonical or self.restart:
                self.get_canonical_representation(ham, wfs, dens)
            else:
                self.dm_helper.set_reference_orbitals(wfs, self.n_dim)
                for kpt in wfs.kpt_u:
                    u = self.kpointval(kpt)
                    self.a_vec_u[u] = np.zeros_like(self.a_vec_u[u])

            # Erase memory of search direction algorithm
            self.searchdir_algo.reset()

    def get_preconditioning(self, wfs, use_prec, make_pd=False):

        if not use_prec:
            return None

        if self.searchdir_algo.name == 'l-bfgs-p':
            beta0 = self.searchdir_algo.beta_0
            gamma = 0.25
        else:
            beta0 = 1.0
            gamma = 0.0

        counter = self.update_precond_counter
        precond = {}
        for kpt in wfs.kpt_u:
            k = self.kpointval(kpt)
            w = kpt.weight / (3.0 - wfs.nspins)
            if self.iters % counter == 0 or self.iters == 1:
                self.hess[k] = self.get_hessian(kpt)
                if make_pd:
                    if self.dtype == float:
                        self.hess[k] = np.abs(self.hess[k])
                    else:
                        self.hess[k] = np.abs(self.hess[k].real) \
                            + 1.0j * np.abs(self.hess[k].imag)
            hess = self.hess[k]
            precond[k] = np.zeros_like(hess)
            correction = w * gamma * beta0 ** (-1)
            if self.searchdir_algo.name != 'l-bfgs-p':
                correction = np.zeros_like(hess)
                zeros = abs(hess) < 1.0e-4
                correction[zeros] = 1.0
            precond[k] += 1. / ((1 - gamma) * hess.real + correction)
            if self.dtype == complex:
                precond[k] += 1.j / ((1 - gamma) * hess.imag + correction)

        return precond

    def get_hessian(self, kpt):
        """
        Calculate the following diagonal approximation to the Hessian:
        h_{lm, lm} = -2.0 * (eps_n[l] - eps_n[m]) * (f[l] - f[m])
        """

        f_n = kpt.f_n
        eps_n = kpt.eps_n
        u = self.kpointval(kpt)
        il1 = list(self.ind_up[u])

        hess = np.zeros(len(il1[0]), dtype=self.dtype)
        x = 0
        for n, m in zip(*il1):
            df = f_n[n] - f_n[m]
            hess[x] = -2.0 * (eps_n[n] - eps_n[m]) * df
            if abs(hess[x]) < 1.0e-10:
                hess[x] = 0.0
            if self.dtype == complex:
                hess[x] += 1.0j * hess[x]
            x += 1

        return hess

    def get_canonical_representation(self, ham, wfs, dens,
                                     sort_eigenvalues=False):
        """
        Choose canonical orbitals as the orbitals that diagonalize the
        Lagrange matrix. It is probably necessary to do a subspace rotation
        with equally occupied orbitals as the total energy is unitary invariant
        within equally occupied subspaces.
        """

        with wfs.timer('Get canonical representation'):
            for kpt in wfs.kpt_u:
                self.dm_helper.update_to_canonical_orbitals(
                    wfs, ham, kpt, self.update_ref_orbs_canonical,
                    self.restart)

            self._e_entropy = wfs.calculate_occupation_numbers(dens.fixed)
            occ_name = getattr(wfs.occupations, "name", None)
            if occ_name == 'mom':
                if not sort_eigenvalues:
                    self.sort_orbitals_mom(wfs)
                else:
                    self.sort_orbitals(ham, wfs, use_eps=True)
                    not_update = not wfs.occupations.update_numbers
                    fixed_occ = wfs.occupations.use_fixed_occupations
                    if not_update or fixed_occ:
                        wfs.occupations.numbers = \
                            self.initial_occupation_numbers

            self.dm_helper.set_reference_orbitals(wfs, self.n_dim)
            for kpt in wfs.kpt_u:
                u = self.kpointval(kpt)
                self.a_vec_u[u] = np.zeros_like(self.a_vec_u[u])

    def reset(self):

        self.dm_helper = None
        self.error = np.inf
        self.initialized = False
        self.searchdir_algo.reset()

    def sort_orbitals(self, ham, wfs, use_eps=False):
        """
        Sort orbitals according to the eigenvalues or
        the diagonal elements of the Hamiltonian matrix
        """

        with wfs.timer('Sort WFS'):
            for kpt in wfs.kpt_u:
                k = self.kpointval(kpt)
                if use_eps:
                    orbital_energies = kpt.eps_n
                else:
                    orbital_energies = self.dm_helper.orbital_energies(
                        wfs, ham, kpt)
                ind = np.argsort(orbital_energies)
                # check if it is necessary to sort
                x = np.max(abs(ind - np.arange(len(ind))))

                if x > 0:
                    # now sort wfs according to orbital energies
                    self.dm_helper.sort_orbitals(wfs, kpt, ind)
                    kpt.f_n[np.arange(len(ind))] = kpt.f_n[ind]
                    kpt.eps_n[np.arange(len(ind))] = orbital_energies[ind]
                    occ_name = getattr(wfs.occupations, "name", None)
                    if occ_name == 'mom':
                        # OccupationsMOM.numbers needs to be updated after
                        # sorting
                        self.update_mom_numbers(wfs, kpt)
                    if self.constraints:
                        # Identity of the contrained orbitals has changed and
                        # needs to be updated
                        self.constraints[k] = update_constraints(
                            self.constraints[k], list(ind))

    def sort_orbitals_mom(self, wfs):
        """
        Sort orbitals according to the occupation
        numbers so that there are no holes in the
        distribution of occupation numbers
        :return:
        """
        changedocc = False
        for kpt in wfs.kpt_u:
            k = self.kpointval(kpt)
            occupied = kpt.f_n > 1.0e-10
            n_occ = len(kpt.f_n[occupied])
            if n_occ == 0.0:
                continue
            if np.min(kpt.f_n[:n_occ]) == 0:
                ind_occ = np.argwhere(occupied)
                ind_unocc = np.argwhere(~occupied)
                ind = np.vstack((ind_occ, ind_unocc))
                ind = np.squeeze(ind)
                self.dm_helper.sort_orbitals(wfs, kpt, ind)
                kpt.f_n = kpt.f_n[ind]
                kpt.eps_n = kpt.eps_n[ind]

                # OccupationsMOM.numbers needs to be updated after sorting
                self.update_mom_numbers(wfs, kpt)

                if self.constraints:
                    # Identities of the contrained orbitals have changed and
                    # needs to be updated
                    self.constraints[k] = update_constraints(
                        self.constraints[k], list(ind))

                changedocc = True

        return changedocc

    def todict(self):

        ret = {'name': self.name,
               'searchdir_algo': self.searchdir_algo.todict(),
               'linesearch_algo': self.line_search.todict(),
               'localizationtype': self.localizationtype,
               'update_ref_orbs_counter': self.update_ref_orbs_counter,
               'update_precond_counter': self.update_precond_counter,
               'use_prec': self.use_prec,
               'matrix_exp': self.matrix_exp,
               'representation': self.representation,
               'functional': self.func.todict(),
               'orthonormalization': self.orthonormalization,
               'constraints': self.constraints
               }
        if self.gmf:
            ret['partial_diagonalizer'] = \
                self.searchdir_algo.partial_diagonalizer.todict()
        return ret

    def rotate_wavefunctions(self, wfs, a_vec_u, n_dim, c_ref):

        """
        Apply unitary transformation U = exp(A) to
        the orbitals c_ref

        :param wfs:
        :param a_vec_u:
        :param n_dim:
        :param c_ref:
        :return:
        """

        with wfs.timer('Unitary rotation'):
            for kpt in wfs.kpt_u:
                k = self.kpointval(kpt)
                if n_dim[k] == 0:
                    continue

                u_nn = self.get_exponential_matrix_kpt(wfs, kpt,
                                                       a_vec_u,
                                                       n_dim)

                self.dm_helper.appy_transformation_kpt(
                    wfs, u_nn.T, kpt, c_ref[k], False, False)

                with wfs.timer('Calculate projections'):
                    self.dm_helper.update_projections(wfs, kpt)

    def get_exponential_matrix_kpt(self, wfs, kpt, a_vec_u, n_dim):
        """
        Get unitary matrix U as the exponential of a skew-Hermitian
        matrix A (U = exp(A))
        """

        k = self.kpointval(kpt)

        if self.gd.comm.rank == 0:
            if self.matrix_exp == 'egdecomp-u-invar' and \
                    self.representation == 'u-invar':
                n_occ = get_n_occ(kpt)
                n_v = n_dim[k] - n_occ
                a_mat = a_vec_u[k].reshape(n_occ, n_v)
            else:
                a_mat = vec2skewmat(a_vec_u[k], n_dim[k],
                                    self.ind_up[k], self.dtype)

            if self.matrix_exp == 'pade-approx':
                # this function takes a lot of memory
                # for large matrices... what can we do?
                with wfs.timer('Pade Approximants'):
                    u_nn = np.ascontiguousarray(expm(a_mat))
            elif self.matrix_exp == 'egdecomp':
                # this method is based on diagonalization
                with wfs.timer('Eigendecomposition'):
                    u_nn, evecs, evals = expm_ed(a_mat, evalevec=True)
            elif self.matrix_exp == 'egdecomp-u-invar':
                with wfs.timer('Eigendecomposition'):
                    u_nn = expm_ed_unit_inv(a_mat, oo_vo_blockonly=False)

        with wfs.timer('Broadcast u_nn'):
            if self.gd.comm.rank != 0:
                u_nn = np.zeros(shape=(n_dim[k], n_dim[k]),
                                dtype=wfs.dtype)
            self.gd.comm.broadcast(u_nn, 0)

        if self.matrix_exp == 'egdecomp':
            with wfs.timer('Broadcast evecs and evals'):
                if self.gd.comm.rank != 0:
                    evecs = np.zeros(shape=(n_dim[k], n_dim[k]),
                                     dtype=complex)
                    evals = np.zeros(shape=n_dim[k],
                                     dtype=float)
                self.gd.comm.broadcast(evecs, 0)
                self.gd.comm.broadcast(evals, 0)
                self.evecs[k], self.evals[k] = evecs, evals

        return u_nn

    def check_assertions(self, wfs, dens):

        assert dens.mixer.driver.basemixerclass.name == 'no-mixing', \
            'Please, use: mixer={\'backend\': \'no-mixing\'}'
        if wfs.occupations.name != 'mom':
            errormsg = \
                'Please, use occupations={\'name\': \'fixed-uniform\'}'
            assert wfs.occupations.name == 'fixed-uniform', errormsg

    def initialize_mom(self, wfs, dens):
        # Reinitialize the MOM reference orbitals
        # after orthogonalization/localization
        wfs.occupations.initialize_reference_orbitals()
        wfs.calculate_occupation_numbers(dens.fixed)
        self.sort_orbitals_mom(wfs)

    def check_mom(self, wfs, dens):
        occ_name = getattr(wfs.occupations, "name", None)
        if occ_name == 'mom':
            self._e_entropy = wfs.calculate_occupation_numbers(dens.fixed)
            self.restart = self.sort_orbitals_mom(wfs)

    def update_mom_numbers(self, wfs, kpt):
        if wfs.collinear and wfs.nspins == 1:
            degeneracy = 2
        else:
            degeneracy = 1
        wfs.occupations.numbers[kpt.s] = \
            kpt.f_n / (kpt.weightk * degeneracy)

    def randomize_orbitals_kpt(self, wfs, kpt):
        """
        Add random noise to orbitals but keep them orthonormal
        """
        nst = self.nbands
        wt = kpt.weight * 0.01
        arand = wt * random_a((nst, nst), wfs.dtype)
        arand = arand - arand.T.conj()
        wfs.gd.comm.broadcast(arand, 0)
        self.dm_helper.appy_transformation_kpt(wfs, expm(arand), kpt)

    def calculate_hamiltonian_matrix(self, hamiltonian, wfs, kpt):
        H_MM = self.dm_helper.calculate_hamiltonian_matrix(
            hamiltonian, wfs, kpt)
        return H_MM

    def kpointval(self, kpt):
        return self.nkpts * kpt.s + kpt.q

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, e):
        self._error = e


def get_n_occ(kpt):
    """
    Get number of occupied orbitals
    """
    return len(kpt.f_n) - np.searchsorted(kpt.f_n[::-1], 1e-10)


def random_a(shape, dtype):

    a = np.random.random_sample(shape)
    if dtype == complex:
        a = a.astype(complex)
        a += 1.0j * np.random.random_sample(shape)

    return a


def vec2skewmat(a_vec, dim, ind_up, dtype):

    a_mat = np.zeros(shape=(dim, dim), dtype=dtype)
    a_mat[ind_up] = a_vec
    a_mat -= a_mat.T.conj()
    np.fill_diagonal(a_mat, a_mat.diagonal() * 0.5)
    return a_mat


def convert_constraints(constraints, n_dim, n_occ, representation):
    """
    Parses and checks the user input of constraints. If constraints are passed
    as a list of a single orbital index all pairs of orbitals involving this
    index are added to the constraints.

    :param constraints: List of constraints for one K-point
    :param n_dim:
    :param n_occ:
    :param representation: Unitary invariant, sparse or full representation
                           determining the electronic degrees of freedom that
                           need to be constrained

    :return: Converted list of constraints
    """

    new = constraints.copy()
    for con in constraints:
        assert isinstance(con, list) or isinstance(con, int), \
            'Check constraints.'
        if isinstance(con, list):
            assert len(con) < 3, 'Check constraints.'
            if len(con) == 1:
                con = con[0]  # List of single index
            else:
                # Make first index always smaller than second index
                if representation != 'full' and con[1] < con[0]:
                    temp = deepcopy(con[0])
                    con[0] = deepcopy(con[1])
                    con[1] = temp
                check_indices(
                    con[0], con[1], n_dim, n_occ, representation)
                continue
        # Add all pairs containing the single index
        if isinstance(con, int):
            new += find_all_pairs(con, n_dim, n_occ, representation)

    # Delete all list containing a single index
    done = False
    while not done:
        done = True
        for i in range(len(new)):
            if len(new[i]) < 2:
                del new[i]
                done = False
                break
    return new


def check_indices(ind1, ind2, n_dim, n_occ, representation):
    """
    Makes sure the user input for the constraints makes sense.
    """

    assert ind1 != ind2, 'Check constraints.'
    if representation == 'full':
        assert ind1 < n_dim and ind2 < n_dim, 'Check constraints.'
    elif representation == 'sparse':
        assert ind1 < n_occ and ind2 < n_dim, 'Check constraints.'
    elif representation == 'u-invar':
        assert ind1 < n_occ and ind2 >= n_occ and ind2 < n_dim, \
            'Check constraints.'


def find_all_pairs(ind, n_dim, n_occ, representation):
    """
    Creates a list of all orbital pairs corresponding to degrees of freedom of
    the system containing an orbital index.

    :param ind: The orbital index
    :param n_dim:
    :param n_occ:
    :param representation: Unitary invariant, sparse or full, defining what
                           index pairs correspond to degrees of freedom of the
                           system
    """

    pairs = []
    if representation == 'u-invar':
        # Only ov rotations are degrees of freedom
        if ind < n_occ:
            for i in range(n_occ, n_dim):
                pairs.append([ind, i])
        else:
            for i in range(n_occ):
                pairs.append([i, ind])
    else:
        if (ind < n_occ and representation == 'sparse') \
                or representation == 'full':
            # oo and ov rotations are degrees of freedom
            for i in range(n_dim):
                if i == ind:
                    continue
                pairs.append([i, ind] if i < ind else [ind, i])
                if representation == 'full':
                    # The orbital rotation matrix is not assumed to be
                    # antihermitian, so the reverse order of indices must be
                    # added as a second constraint
                    pairs.append([ind, i] if i < ind else [i, ind])
        else:
            for i in range(n_occ):
                pairs.append([i, ind])
    return pairs


def update_constraints(constraints, ind):
    """
    Change the constraint indices to match a new indexation, e.g. due to
    sorting the orbitals

    :param constraints: The list of constraints for one K-point
    :param ind: List containing information about the change in indexation
    """

    new = deepcopy(constraints)
    for i in range(len(constraints)):
        for k in range(len(constraints[i])):
            new[i][k] = ind.index(constraints[i][k])
    return new



import hashlib

import numpy as np

from ase.units import Hartree, Bohr


def L_to_lm(L):
    """Convert L index to (l, m) index."""
    l = int(np.sqrt(L))
    m = L - l**2 - l
    return l, m


def lm_to_L(l, m):
    """Convert (l, m) index to L index."""
    return l**2 + l + m


def split_formula(formula):
    """Count elements in a chemical formula.

    E.g. split_formula('C2H3Mg') -> ['C', 'C', 'H', 'H', 'H', 'Mg']
    """
    res = []
    for c in formula:
        if c.isupper():
            res.append(c)
        elif c.islower():
            res[-1] += c
        else:
            res.extend([res[-1]] * (eval(c) - 1))
    return res


def construct_reciprocal(gd, q_c=None):
    """Construct the reciprocal lattice from ``GridDescriptor`` instance.

    The generated reciprocal lattice has lattice vectors corresponding to the
    real-space lattice defined in the input grid. Note that it is the squared
    length of the reciprocal lattice vectors that are returned.

    The ordering of the reciprocal lattice agrees with the one typically used
    in fft algorithms, i.e. positive k-values followed by negative.

    Note that the G=(0,0,0) entry is set to one instead of zero. This
    bit should probably be moved somewhere else ...

    Parameters
    ----------
    q_c: ndarray
        Offset for the reciprocal lattice vectors (in scaled coordinates of the
        reciprocal lattice vectors, i.e. array with index ``c``). When
        specified, the returned array contains the values of (q+G)^2 where G
        denotes the reciprocal lattice vectors.

    """

    assert gd.pbc_c.all(), 'Works only with periodic boundary conditions!'

    q_c = np.zeros((3, 1), dtype=float) if q_c is None else q_c.reshape((3, 1))

    # Calculate reciprocal lattice vectors
    N_c1 = gd.N_c[:, None]
    i_cq = np.indices(gd.n_c, dtype=float).reshape((3, -1))  # offsets....
    i_cq += gd.beg_c[:, None]
    i_cq += N_c1 // 2
    i_cq %= N_c1
    i_cq -= N_c1 // 2

    i_cq += q_c

    # Convert from scaled to absolute coordinates
    B_vc = 2.0 * np.pi * gd.icell_cv.T
    k_vq = np.dot(B_vc, i_cq)

    k_vq *= k_vq
    k2_Q = k_vq.sum(axis=0).reshape(gd.n_c)

    # Avoid future divide-by-zero by setting k2_Q[G=(0,0,0)] = 1.0 if needed
    if k2_Q[0, 0, 0] < 1e-10:
        k2_Q[0, 0, 0] = 1.0           # Only make sense iff
        assert gd.comm.rank == 0      # * on rank 0 (G=(0,0,0) is only there)
        assert abs(q_c).sum() < 1e-8  # * q_c is (almost) zero

    assert k2_Q.min() > 0.0       # Now there should be no zero left

    # Determine N^3
    #
    # Why do we need to calculate and return this?  The caller already
    # has access to gd and thus knows how many points there are.
    N3 = gd.n_c[0] * gd.n_c[1] * gd.n_c[2]
    return k2_Q, N3


def coordinates(gd, origin=None, tiny=1e-12):
    """Constructs and returns matrices containing cartesian coordinates,
       and the square of the distance from the origin.

       The origin can be given explicitely (in Bohr units, not Anstroms).
       Otherwise the origin is placed in the center of the box described
       by the given grid-descriptor 'gd'.
    """

    if origin is None:
        origin = 0.5 * gd.cell_cv.sum(0)
    r0_v = np.array(origin)

    r_vG = gd.get_grid_point_distance_vectors(r0_v)
    r2_G = np.sum(r_vG**2, axis=0)
    # Remove singularity at origin and replace with small number
    r2_G = np.where(r2_G < tiny, tiny, r2_G)

    # Return r^2 matrix
    return r_vG, r2_G


def pick(a_ix, i):
    """Take integer index of a, or a linear combination of the elements of a"""
    if isinstance(i, int):
        return a_ix[i]
    shape = a_ix.shape
    a_x = np.dot(i, a_ix[:].reshape(shape[0], -1))
    return a_x.reshape(shape[1:])


def dagger(a, copy=True):
    """Return Hermitian conjugate of input

    If copy is False, the original array might be overwritten. This is faster,
    but use with care.
    """
    if copy:
        return np.conj(a.T)
    else:
        a = a.T
        if a.dtype == complex:
            a.imag *= -1
        return a


def project(a, b):
    """Return the projection of b onto a."""
    return a * (np.dot(a.conj(), b) / np.linalg.norm(a))


def normalize(U):
    """Normalize columns of U."""
    for col in U.T:
        col /= np.linalg.norm(col)


def get_matrix_index(ind1, ind2=None):
    if ind2 is None:
        dim1 = len(ind1)
        return np.resize(ind1, (dim1, dim1))
    else:
        dim1 = len(ind1)
        dim2 = len(ind2)
    return np.resize(ind1, (dim2, dim1)).T, np.resize(ind2, (dim1, dim2))


def gram_schmidt(U):
    """Orthonormalize columns of U according to the Gram-Schmidt procedure."""
    for i, col in enumerate(U.T):
        for col2 in U.T[:i]:
            col -= col2 * np.dot(col2.conj(), col)
        col /= np.linalg.norm(col)


def lowdin(U, S=None):
    """Orthonormalize columns of U according to the Lowdin procedure.

    If the overlap matrix is know, it can be specified in S.
    """
    if S is None:
        S = np.dot(dagger(U), U)
    eig, rot = np.linalg.eigh(S)
    rot = np.dot(rot / np.sqrt(eig), dagger(rot))
    U[:] = np.dot(U, rot)


def lowdin_svd(U):
    """Orthogonalize according to the Lowdin procedure
       using singular value decomposition.

       U is an N x M matrix containing M vectors as its columns.
    """
    Z, D, V = np.linalg.svd(U, full_matrices=0)
    return np.dot(Z, V)


def symmetrize(matrix):
    """Symmetrize input matrix."""
    np.add(dagger(matrix), matrix, matrix)
    np.multiply(.5, matrix, matrix)
    return matrix


def tri2full(H_nn, UL='L', map=np.conj):
    """Fill in values of hermitian or symmetric matrix.

    Fill values in lower or upper triangle of H_nn based on the opposite
    triangle, such that the resulting matrix is symmetric/hermitian.

    UL='U' will copy (conjugated) values from upper triangle into the
    lower triangle.

    UL='L' will copy (conjugated) values from lower triangle into the
    upper triangle.

    The map parameter can be used to specify a different operation than
    conjugation, which should work on 1D arrays.  Example::

      def antihermitian(src, dst):
            np.conj(-src, dst)

      tri2full(H_nn, map=antihermitian)

    """
    N, tmp = H_nn.shape
    assert N == tmp, 'Matrix must be square'
    if UL != 'L':
        H_nn = H_nn.T

    for n in range(N - 1):
        map(H_nn[n + 1:, n], H_nn[n, n + 1:])


def apply_subspace_mask(H_nn, f_n):
    """Uncouple occupied and unoccupied subspaces.

    This method forces the H_nn matrix into a block-diagonal form
    in the occupied and unoccupied states respectively.
    """
    occ = 0
    nbands = len(f_n)
    while occ < nbands and f_n[occ] > 1e-3:
        occ += 1
    H_nn[occ:, :occ] = H_nn[:occ, occ:] = 0


def cutoff2gridspacing(E):
    """Convert planewave energy cutoff to a real-space gridspacing."""
    return np.pi / np.sqrt(2 * E / Hartree) * Bohr


def gridspacing2cutoff(h):
    """Convert real-space gridspacing to planewave energy cutoff."""
    # In Hartree units, E = k^2 / 2, where k_max is approx. given by pi / h
    # See PRB, Vol 54, 14362 (1996)
    return 0.5 * (np.pi * Bohr / h)**2 * Hartree


def tridiag(a, b, c, r, u):
    """Solve linear system with tridiagonal coefficient matrix.

    a is the lower band, b is the diagonal, c is the upper band, and
    r is the right hand side.
    The solution is returned in u.


    [b1 c1  0  ...            ] [u1]   [r1]
    [a1 b2 c2 0 ...           ] [ :]   [ :]
    [ 0 a2 b3 c3 0 ...        ] [  ] = [  ]
    [                         ] [  ]   [  ]
    [     ... 0 an-2 bn-1 cn-1] [ :]   [ :]
    [          ... 0 an-1 bn  ] [un]   [rn]
    """
    n = len(b)
    tmp = np.zeros(n - 1)  # necessary temporary array
    if b[0] == 0:
        raise RuntimeError('System is effectively order N-1')

    beta = b[0]
    u[0] = r[0] / beta
    for i in range(1, n):
        # Decompose and forward substitution
        tmp[i - 1] = c[i - 1] / beta
        beta = b[i] - a[i - 1] * tmp[i - 1]
        if beta == 0:
            raise RuntimeError('Method failure')
        u[i] = (r[i] - a[i - 1] * u[i - 1]) / beta

    for i in range(n - 1, 0, -1):
        # Backward substitution
        u[i - 1] -= tmp[i - 1] * u[i]


def signtrim(data, decimals=None):
    """Trim off the sign of potential zeros, usually occurring after round.

    data is the ndarray holding NumPy data to round and trim.
    decimals is an integer specifying how many decimals to round to.
    """
    if decimals is not None:
        data = data.round(decimals)  # np.round is buggy because -0 != 0

    shape = data.shape
    data = data.reshape(-1)

    if data.dtype == complex:
        i = np.argwhere(np.sign(data.real) == 0).ravel()
        j = np.argwhere(np.sign(data.imag) == 0).ravel()
        data.real[i] = 0
        data.imag[j] = 0
    else:
        i = np.argwhere(np.sign(data) == 0).ravel()
        data[i] = 0

    return data.reshape(shape)


def md5_array(data, numeric=False):
    """Create MD5 hex digest from NumPy array.

    Optionally, will cast the 128 bit hexadecimal hash to a 64 bit integer.

    Warning: For floating point types, only bitwise identical data will
    produce matching MD5 fingerprints, so do not attempt to match sets
    of nearly identical data by rounding off beforehand.

    Example:

     >>> data = np.linspace(0, np.pi, 1000000)
     >>> eps = 1e-6
     >>> a = md5_array(data.round(3))
     >>> b = md5_array((data + eps).round(3))
     >>> a == b
     False

    This is due to the inexact nature of the floating point representation.
    """

    if not isinstance(data, np.ndarray):
        data = np.asarray(data)

    # Only accepts float,complex,int,bool,...
    if (not np.issubdtype(data.dtype, np.number) and
        data.dtype not in [bool, np.bool, np.bool_]):
        raise TypeError('MD5 hex digest only accepts numeric/boolean arrays.')

    datahash = hashlib.md5(data.tobytes())

    if numeric:
        def xor(a, b):
            return chr(ord(a) ^ ord(b))  # bitwise xor on 2 bytes -> 1 byte

        sbuf128 = datahash.digest()
        sbuf64 = ''.join([xor(a, b)
                          for a, b in zip(sbuf128[::2], sbuf128[1::2])])
        return np.fromstring(sbuf64, np.int64).item()
    else:
        return datahash.hexdigest()


def split_nodes(length, parrank, parsize):
    """Split length over nodes.

    Divide length into parsize even sized chunks, and return the start/end
    indices of the parrank'th chunk.
    """
    if parsize == 1:
        return 0, length
    pernode = int(round(length / float(parsize)))
    if parrank == parsize - 1:
        return parrank * pernode, length
    return parrank * pernode, (parrank + 1) * pernode


class Spline:
    def __init__(self, xi, yi, leftderiv=None, rightderiv=None):
        """Cubic spline approximation class.

        xi, yi specifies the known data points.

        leftderiv and rightderiv specifies the first derivative on the
        boundaries. If set to None, the second derivative is set to zero.

        Example usage::

          >>> xi = np.arange(.1, 5, .5)    # known data points
          >>> yi = np.cos(xi)              # known data points
          >>> sp = Spline(xi, yi)       # make spline
          >>> x = np.arange(-.5, 5.5, .05) # points to interpolate to
          >>> y = sp(x)  # get spline value on an entire list
          >>> y2 = sp(4) # get spline value at a single point

        Based on 'Numerical recipes in c'
        """
        self.xy = (xi, yi)
        N = len(xi)
        self.ypp = u = np.zeros(N)  # the second derivatives y''
        tmp = np.zeros(N - 1)

        # Set left boundary condition
        if leftderiv is None:  # natural spline - second derivative is zero
            tmp[0] = u[0] = 0.0
        else:  # clamped spline - first derivative is fixed
            tmp[0] = 3 / (xi[1] - xi[0]) * (
                (yi[1] - yi[0]) / (xi[1] - xi[0]) - leftderiv)
            u[0] = -.5

        for i in range(1, N - 1):
            sig = (xi[i] - xi[i - 1]) / (xi[i + 1] - xi[i - 1])
            p = sig * u[i - 1] + 2
            u[i] = (sig - 1) / p
            tmp[i] = (yi[i + 1] - yi[i]) / (xi[i + 1] - xi[i]) - \
                     (yi[i] - yi[i - 1]) / (xi[i] - xi[i - 1])
            tmp[i] = (6 * tmp[i] /
                      (xi[i + 1] - xi[i - 1]) - sig * tmp[i - 1]) / p

        # Set right boundary condition
        if rightderiv is None:  # natural spline - second derivative is zero
            qn = tmpn = 0.0
        else:  # clamped spline - first derivative is fixed
            qn = .5
            tmpn = 3 / (xi[N - 1] - xi[N - 2]) * (
                rightderiv - (yi[N - 1] - yi[N - 2]) / (xi[N - 1] - xi[N - 2]))

        u[N - 1] = (tmpn - qn * tmp[N - 2]) / (qn * u[N - 1] + 1)
        for k in range(N - 2, -1, -1):  # backsubstitution step
            u[k] = u[k] * u[k + 1] + tmp[k]

    def __call__(self, x):
        """Evaluate spline for each point in input argument.

        The value in point x[i-1] < x <= x[i] is determined by::

                                    ''       ''
          y(x) = a y    + b y  + c y    + d y
                    i-1      i      i-1      i

        """
        x = np.array(x, float)
        if x.ndim == 0:
            x.shape = (1,)
        y = np.zeros_like(x)
        xi, yi = self.xy

        i = None
        for j, xval in enumerate(x):
            i = self.locate(xval, i)
            h = xi[i] - xi[i - 1]
            a = (xi[i] - xval) / h
            b = 1. - a
            c = (a**3 - a) * h**2 / 6.
            d = (b**3 - b) * h**2 / 6.
            y[j] = (a * yi[i - 1] + b * yi[i] +
                    c * self.ypp[i - 1] + d * self.ypp[i])
        return y

    def locate(self, x, guess=None):
        """return i such that x[i-1] < x <= xi[i]

        1 or len(xi) - 1 is returned if x is outside list range.
        """
        xi = self.xy[0]
        if x <= xi[0]:
            return 1
        elif x > xi[-1]:
            return len(xi) - 1
        elif guess and xi[guess - 1] < x <= xi[guess]:
            return guess
        else:
            return np.searchsorted(xi, x)



import numpy as np

from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.projections import Projections
from gpaw.utilities.partition import AtomPartition
from gpaw.wavefunctions.arrays import PlaneWaveExpansionWaveFunctions
from gpaw.pw.descriptor import PWDescriptor


class KPoint:
    def __init__(self,
                 proj,  # projections
                 f_n,  # occupations numbers between 0 and 1
                 k_c,  # k-vector in units of reciprocal cell
                 weight,  # weight of k-point
                 dPdR_aniv=[]):
        self.proj = proj
        self.f_n = f_n
        self.k_c = k_c
        self.weight = weight
        self.dPdR_aniv = dPdR_aniv


class PWKPoint(KPoint):
    def __init__(self, psit, *args):  # plane-wave expansion of wfs
        self.psit = psit
        KPoint.__init__(self, *args)


class RSKPoint(KPoint):
    def __init__(self, u_nR, *args):
        self.u_nR = u_nR
        KPoint.__init__(self, *args)


def to_real_space(psit, na=0, nb=None):
    pd = psit.pd
    comm = pd.comm
    S = comm.size
    q = psit.kpt
    nbands = len(psit.array)
    nb = nb or nbands
    u_nR = pd.gd.empty(nbands, pd.dtype, global_array=True)
    for n1 in range(0, nbands, S):
        n2 = min(n1 + S, nbands)
        u_G = pd.alltoall1(psit.array[n1:n2], q)
        if u_G is not None:
            n = n1 + comm.rank
            u_nR[n] = pd.ifft(u_G, local=True, safe=False, q=q)
        for n in range(n1, n2):
            comm.broadcast(u_nR[n], n - n1)

    return u_nR[na:nb]


def get_kpt(wfs, k, spin, n1, n2):
    k_c = wfs.kd.ibzk_kc[k]
    weight = wfs.kd.weight_k[k]

    if wfs.world.size == wfs.gd.comm.size:
        # Easy:
        kpt = wfs.kpt_qs[k][spin]
        psit = kpt.psit.view(n1, n2)
        proj = kpt.projections.view(n1, n2)
        f_n = kpt.f_n[n1:n2]
    else:
        # Need to redistribute things:
        gd = wfs.gd.new_descriptor(comm=wfs.world)
        kd = KPointDescriptor([k_c])
        pd = PWDescriptor(wfs.ecut, gd, wfs.dtype, kd, wfs.fftwflags)
        psit = PlaneWaveExpansionWaveFunctions(n2 - n1,
                                               pd,
                                               dtype=wfs.dtype,
                                               spin=spin)
        for n in range(n1, n2):
            psit_G = wfs.get_wave_function_array(n, k, spin,
                                                 realspace=False, cut=False)
            if isinstance(psit_G, float):
                psit_G = None
            else:
                psit_G = psit_G[:pd.ngmax]
            psit._distribute(psit_G, psit.array[n - n1])

        P_nI = wfs.collect_projections(k, spin)
        if wfs.world.rank == 0:
            P_nI = P_nI[n1:n2]
        natoms = len(wfs.setups)
        rank_a = np.zeros(natoms, int)
        atom_partition = AtomPartition(wfs.world, rank_a)
        nproj_a = [setup.ni for setup in wfs.setups]
        proj = Projections(n2 - n1,
                           nproj_a,
                           atom_partition,
                           spin=spin,
                           dtype=wfs.dtype,
                           data=P_nI)

        rank_a = np.linspace(0, wfs.world.size, len(wfs.setups),
                             endpoint=False).astype(int)
        atom_partition = AtomPartition(wfs.world, rank_a)
        proj = proj.redist(atom_partition)

        f_n = wfs.collect_occupations(k, spin)
        if wfs.world.rank != 0:
            f_n = np.empty(n2 - n1)
        else:
            f_n = f_n[n1:n2]
        wfs.world.broadcast(f_n, 0)

    f_n = f_n / (weight * (2 // wfs.nspins))  # scale to [0, 1]

    return PWKPoint(psit, proj, f_n, k_c, weight)



"""PW-mode stress tensor calculation."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from gpaw.core.atom_arrays import AtomArrays
from gpaw.gpu import synchronize, as_xp
from gpaw.new.calculation import DFTState
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
from gpaw.typing import Array2D
from gpaw.core import PWArray
if TYPE_CHECKING:
    from gpaw.new.pw.pot_calc import PlaneWavePotentialCalculator


def calculate_stress(pot_calc: PlaneWavePotentialCalculator,
                     state: DFTState,
                     vt_g: PWArray,
                     nt_g: PWArray,
                     dedtaut_g: PWArray | None) -> Array2D:
    """Calculate symmetrized stress tensor."""
    ibzwfs = state.ibzwfs
    density = state.density
    potential = state.potential
    comm = ibzwfs.comm
    xp = density.nt_sR.xp
    dom = density.nt_sR.desc

    ibzwfs.make_sure_wfs_are_read_from_gpw_file()
    s_vv = get_wfs_stress(ibzwfs, potential.dH_asii)
    s_vv += pot_calc.xc.stress_contribution(state, pot_calc.interpolate)

    if state.ibzwfs.kpt_comm.rank == 0 and state.ibzwfs.band_comm.rank == 0:
        vHt_h = potential.vHt_x
        assert vHt_h is not None
        pw = vHt_h.desc
        G_Gv = xp.asarray(pw.G_plus_k_Gv)
        vHt2_hz = vHt_h.data.view(float).reshape((len(G_Gv), 2))**2
        s_vv += (xp.einsum('Gz, Gv, Gw -> vw', vHt2_hz, G_Gv, G_Gv) *
                 pw.dv / (2 * np.pi))
        Q_aL = density.calculate_compensation_charge_coefficients()
        s_vv += pot_calc.ghat_aLh.stress_contribution(vHt_h, Q_aL)
        if state.ibzwfs.domain_comm.rank == 0:
            s_vv -= xp.eye(3) * potential.energies['stress']
        s_vv += pot_calc.vbar_ag.stress_contribution(nt_g)
        s_vv += density.nct_aX.stress_contribution(vt_g)

        if dedtaut_g is not None:
            s_vv += density.tauct_aX.stress_contribution(dedtaut_g)

    s_vv = as_xp(s_vv, np)

    if xp is not np:
        synchronize()
    comm.sum(s_vv, 0)

    vol = dom.volume
    s_vv = 0.5 / vol * (s_vv + s_vv.T)

    # Symmetrize:
    sigma_vv = np.zeros((3, 3))
    cell_cv = dom.cell_cv
    icell_cv = dom.icell
    rotation_scc = ibzwfs.ibz.symmetries.rotation_scc
    for U_cc in rotation_scc:
        M_vv = (icell_cv.T @ (U_cc @ cell_cv)).T
        sigma_vv += M_vv.T @ s_vv @ M_vv
    sigma_vv /= len(rotation_scc)

    # Make sure all agree on the result (redundant calculation on
    # different cores involving BLAS might give slightly different
    # results):
    comm.broadcast(sigma_vv, 0)
    return sigma_vv


def get_wfs_stress(ibzwfs: IBZWaveFunctions,
                   dH_asii: AtomArrays) -> Array2D:
    xp = ibzwfs.xp
    sigma_vv = xp.zeros((3, 3))
    for wfs in ibzwfs:
        assert isinstance(wfs, PWFDWaveFunctions)
        occ_n = xp.asarray(wfs.weight * wfs.spin_degeneracy * wfs.myocc_n)
        sigma_vv += get_kinetic_stress(wfs, occ_n)
        sigma_vv += get_paw_stress(wfs, dH_asii, occ_n)
    return sigma_vv


def get_kinetic_stress(wfs: PWFDWaveFunctions,
                       occ_n) -> Array2D:
    psit_nG = wfs.psit_nX
    pw = psit_nG.desc
    xp = psit_nG.xp
    psit_nGz = psit_nG.data.view(float).reshape(psit_nG.data.shape + (2,))
    psit2_G = xp.einsum('n, nGz, nGz -> G', occ_n, psit_nGz, psit_nGz)
    Gk_Gv = xp.asarray(pw.G_plus_k_Gv)
    sigma_vv = xp.einsum('G, Gv, Gw -> vw', psit2_G, Gk_Gv, Gk_Gv)
    x = pw.dv
    if pw.dtype == float:
        x *= 2
    return -x * sigma_vv


def get_paw_stress(wfs: PWFDWaveFunctions,
                   dH_asii: AtomArrays,
                   occ_n) -> Array2D:
    xp = wfs.xp
    eig_n1 = xp.asarray(wfs.myeig_n[:, None])
    a_ani = {}
    s = 0.0
    for a, P_ni in wfs.P_ani.items():
        Pf_ni = P_ni * occ_n[:, None]
        dH_ii = dH_asii[a][wfs.spin]
        dS_ii = xp.asarray(wfs.setups[a].dO_ii)
        a_ni = Pf_ni @ dH_ii - Pf_ni * eig_n1 @ dS_ii
        s += xp.vdot(P_ni, a_ni)
        a_ani[a] = 2 * a_ni.conj()
    s_vv = wfs.pt_aiX.stress_contribution(wfs.psit_nX, a_ani)
    return s_vv - float(s.real) * xp.eye(3)



import pytest
from ase import Atoms
from gpaw import GPAW, PW


@pytest.mark.libxc
@pytest.mark.hybrids
def test_exx_double_cell(in_tmp_dir):
    L = 2.6
    a = Atoms('H2',
              [[0, 0, 0], [0.5, 0.5, 0]],
              cell=[L, L, 1],
              pbc=1)
    a.center()

    a.calc = GPAW(
        mode=PW(400),
        symmetry='off',
        kpts={'size': (1, 1, 4), 'gamma': True},
        convergence={'density': 1e-6},
        spinpol=True,
        txt='H2.txt',
        xc='HSE06')
    e1 = a.get_potential_energy()
    eps1 = a.calc.get_eigenvalues(1)[0]
    f1 = a.get_forces()
    assert abs(f1[1, 0] - 9.60644) < 0.0005
    if 0:
        # To check against numeric calculation of the forces, but it takes
        # more time
        from ase.calculators.test import numeric_force
        f1n = numeric_force(a, 1, 0, 0.001)
        assert abs(f1[1, 0] - f1n) < 0.0005

    a *= (1, 1, 2)
    a.calc = GPAW(
        mode=PW(400),
        kpts={'size': (1, 1, 2), 'gamma': True},
        symmetry='off',
        convergence={'density': 1e-6},
        txt='H4.txt',
        xc='HSE06')
    e2 = a.get_potential_energy()
    eps2 = a.calc.get_eigenvalues(0)[0]
    f2 = a.get_forces()

    f2[:2] -= f1
    f2[2:] -= f1

    assert abs(e2 - 2 * e1) < 0.002
    assert abs(eps1 - eps2) < 0.001
    assert abs(f2).max() < 0.00085


if __name__ == '__main__':
    from cProfile import Profile
    prof = Profile()
    prof.enable()
    test_exx_double_cell(1)
    prof.disable()
    from gpaw.mpi import rank, size
    prof.dump_stats(f'prof-{size}.{rank}')



import pytest
from gpaw.mpi import world
from ase import Atoms
from gpaw import GPAW
from gpaw.test import equal

from gpaw.utilities.kspot import AllElectronPotential
pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


@pytest.mark.later
def test_muffintinpot(in_tmp_dir):
    if 1:
        be = Atoms(symbols='Be', positions=[(0, 0, 0)])
        be.center(vacuum=5)
        calc = GPAW(mode='fd',
                    gpts=(64, 64, 64),
                    xc='LDA',
                    nbands=1)  # 0.1 required for accuracy
        be.calc = calc
        e = be.get_potential_energy()

        energy_tolerance = 0.001
        equal(e, 0.00246471, energy_tolerance)

    # be, calc = restart("be.gpw")
    AllElectronPotential(calc).write_spherical_ks_potentials('bepot.txt')
    f = open('bepot.txt')
    lines = f.readlines()
    f.close()
    mmax = 0
    for l in lines:
        mmax = max(abs(eval(l.split(' ')[3])), mmax)

    print("Max error: ", mmax)
    assert mmax < 0.009



import pytest

from ase import Atoms

from gpaw import GPAW
from gpaw.external import StepPotentialz


@pytest.mark.later
def test_He():
    a = 3
    b = 3 * a
    atoms = Atoms('He', positions=[(a, a, a)],
                  cell=(2 * a, 2 * a, b))

    c00 = GPAW(mode='fd', charge=1)
    atoms.calc = c00
    atoms.get_potential_energy()

    # apply potential where the atom is
    constant = -5
    c10 = GPAW(mode='fd',
               charge=1,
               external=StepPotentialz(b / 2, value_left=constant))
    atoms.calc = c10
    atoms.get_potential_energy()

    # the total energy is shifted for charged systems
    assert c00.get_potential_energy() == pytest.approx(
        c10.get_potential_energy() + constant, 1e-4)

    # apply potential where there is no atom
    c01 = GPAW(mode='fd',
               charge=1,
               external=StepPotentialz(b / 2, value_right=constant))
    atoms.calc = c01
    atoms.get_potential_energy()

    # the total energy should be the same as without the potential
    assert c00.get_potential_energy() == pytest.approx(
        c01.get_potential_energy(), 1e-4)



import pytest


@pytest.mark.response
def test_pair_density_paw_correction():
    import numpy as np
    from gpaw.lfc import LocalizedFunctionsCollection as LFC
    from gpaw.grid_descriptor import GridDescriptor
    from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
    from gpaw.spline import Spline
    from gpaw.response.paw import calculate_pair_density_correction, Setuplet
    # Initialize s, p, d (9 in total) wave and put them on grid
    rc = 2.0
    a = 2.5 * rc
    n = 64
    lmax = 2
    b = 8.0
    m = (lmax + 1)**2
    gd = GridDescriptor([n, n, n], [a, a, a])
    r = np.linspace(0, rc, 200)
    g = np.exp(-(r / rc * b)**2)
    splines = [Spline(l=l, rmax=rc, f_g=g) for l in range(lmax + 1)]
    c = LFC(gd, [splines])
    c.set_positions([(0.5, 0.5, 0.5)])
    psi = gd.zeros(m)
    d0 = c.dict(m)
    if 0 in d0:
        d0[0] = np.identity(m)
    c.add(psi, d0)

    # Calculate on 3d-grid < phi_i | e**(-ik.r) | phi_j >
    R_a = np.array([a / 2, a / 2, a / 2])
    rr = gd.get_grid_point_coordinates()
    for dim in range(3):
        rr[dim] -= R_a[dim]

    k_G = np.array([[0.0, 0.0, 0.0], [1., 0.2, 0.1], [10., 0., 10.]])
    nkpt = k_G.shape[0]

    d0 = np.zeros((nkpt, m, m), dtype=complex)
    for i in range(m):
        for j in range(m):
            for ik in range(nkpt):
                k = k_G[ik]
                # kk = np.sqrt(np.inner(k, k))
                kr = np.inner(k, rr.T).T
                expkr = np.exp(-1j * kr)
                d0[ik, i, j] = gd.integrate(psi[i] * psi[j] * expkr)

    # Calculate on 1d-grid < phi_i | e**(-ik.r) | phi_j >
    rgd = EquidistantRadialGridDescriptor(r[1], len(r))
    g = [np.exp(-(r / rc * b)**2) * r**l for l in range(lmax + 1)]
    l_j = range(lmax + 1)
    rcut_j = [rc] * (lmax + 1)
    d1 = calculate_pair_density_correction(k_G, pawdata=Setuplet(
        rgd=rgd, phi_jg=g, phit_jg=np.zeros_like(g), l_j=l_j, rcut_j=rcut_j))

    assert d0 == pytest.approx(d1, abs=1e-8)



"""Test read/write of restart files between fd and lcao mode"""
import os

import pytest
from ase import Atom, Atoms

from gpaw import GPAW, FermiDirac, restart
from gpaw.test import equal


@pytest.mark.later
def test_lcao_fd2lcao_restart(in_tmp_dir):
    energy_tolerance = 0.001

    if not os.path.isfile('Na4_fd.gpw'):
        # Do grid kpts calculation
        a = 3.31
        atoms = Atoms([Atom('Na', (i * a, 0, 0))
                       for i in range(4)], pbc=(1, 0, 0))
        atoms.center(vacuum=3.5)
        atoms.center(vacuum=a / 2, axis=0)

        calc = GPAW(mode='fd',
                    nbands=-3,
                    h=0.3,
                    setups={'Na': '1'},
                    xc={'name': 'PBE', 'stencil': 1},
                    occupations=FermiDirac(width=0.1),
                    kpts=(3, 1, 1),
                    # basis='dzp',
                    txt='Na4_fd.txt')
        atoms.calc = calc
        etot_fd = atoms.get_potential_energy()
        print('Etot:', etot_fd, 'eV in fd-mode')
        calc.write('Na4_fd.gpw')
        del atoms, calc

        equal(etot_fd, -1.99055, energy_tolerance)

    if os.path.isfile('Na4_fd.gpw'):
        # LCAO calculation based on grid kpts calculation
        atoms, calc = restart('Na4_fd.gpw',
                              # basis='dzp',
                              mode='lcao',
                              txt='Na4_lcao.txt')
        etot_lcao = atoms.get_potential_energy()
        print('Etot:', etot_lcao, 'eV in lcao-mode')
        calc.write('Na4_lcao.gpw')
        del atoms, calc

        equal(etot_lcao, -1.9616, energy_tolerance)



from gpaw.mixer import (MixerSum, SeparateSpinMixerDriver, SpinSumMixerDriver,
                        get_mixer_from_keywords)


def test_mixer_sum_with_spin_paired_calc():
    dct = MixerSum(0.1, 3, 10)
    mixer = get_mixer_from_keywords(pbc=[1, 1, 1], nspins=1, **dct)
    assert isinstance(mixer, SeparateSpinMixerDriver)
    mixer = get_mixer_from_keywords(pbc=[1, 1, 1], nspins=2, **dct)
    assert isinstance(mixer, SpinSumMixerDriver)



from __future__ import annotations

import warnings
from time import ctime
from typing import Union

import numpy as np
from ase.units import Ha

import gpaw
import gpaw.mpi as mpi
from gpaw.bztools import convex_hull_volume
from gpaw.response.chi0_data import (Chi0Data, Chi0BodyData,
                                     Chi0OpticalExtensionData)
from gpaw.response.frequencies import (FrequencyDescriptor,
                                       NonLinearFrequencyDescriptor)
from gpaw.response.pair_functions import SingleQPWDescriptor
from gpaw.response.hilbert import HilbertTransform
from gpaw.response.integrators import (
    Integrand, PointIntegrator, TetrahedronIntegrator)
from gpaw.response import timer
from gpaw.response.pair import KPointPairFactory
from gpaw.response.pw_parallelization import PlaneWaveBlockDistributor
from gpaw.response.symmetry import PWSymmetryAnalyzer
from gpaw.typing import Array1D
from gpaw.utilities.memory import maxrss
from gpaw.response.integrators import (
    HermitianOpticalLimit, HilbertOpticalLimit, OpticalLimit,
    HilbertOpticalLimitTetrahedron,
    Hermitian, Hilbert, HilbertTetrahedron, GenericUpdate)


def find_maximum_frequency(kpt_u, context, nbands=0):
    """Determine the maximum electron-hole pair transition energy."""
    epsmin = 10000.0
    epsmax = -10000.0
    for kpt in kpt_u:
        epsmin = min(epsmin, kpt.eps_n[0])
        epsmax = max(epsmax, kpt.eps_n[nbands - 1])

    context.print('Minimum eigenvalue: %10.3f eV' % (epsmin * Ha),
                  flush=False)
    context.print('Maximum eigenvalue: %10.3f eV' % (epsmax * Ha))

    return epsmax - epsmin


class Chi0Integrand(Integrand):
    def __init__(self, chi0calc, optical, qpd, analyzer, m1, m2):
        self._chi0calc = chi0calc

        # In a normal response calculation, we include transitions from all
        # completely and partially unoccupied bands to range(m1, m2)
        self.n1 = 0
        self.n2 = chi0calc.nocc2
        assert m1 <= m2
        self.m1 = m1
        self.m2 = m2

        self.context = chi0calc.context
        self.kptpair_factory = chi0calc.kptpair_factory
        self.gs = chi0calc.gs

        self.qpd = qpd
        self.analyzer = analyzer
        self.integrationmode = chi0calc.integrationmode
        self.optical = optical

    @timer('Get matrix element')
    def matrix_element(self, k_v, s):
        """Return pair density matrix element for integration.

        A pair density is defined as::

         <snk| e^(-i (q + G) r) |s'mk+q>,

        where s and s' are spins, n and m are band indices, k is
        the kpoint and q is the momentum transfer. For dielectric
        response s'=s, for the transverse magnetic response
        s' is flipped with respect to s.

        Parameters
        ----------
        k_v : ndarray
            Kpoint coordinate in cartesian coordinates.
        s : int
            Spin index.

        If self.optical, then return optical pair densities, that is, the
        head and wings matrix elements indexed by:
        # P = (x, y, v, G1, G2, ...).

        Return
        ------
        n_nmG : ndarray
            Pair densities.
        """

        if self.optical:
            target_method = self._chi0calc.pair_calc.get_optical_pair_density
            out_ngmax = self.qpd.ngmax + 2
        else:
            target_method = self._chi0calc.pair_calc.get_pair_density
            out_ngmax = self.qpd.ngmax

        return self._get_any_matrix_element(
            k_v, s, block=not self.optical,
            target_method=target_method,
        ).reshape(-1, out_ngmax)

    def _get_any_matrix_element(self, k_v, s, block, target_method):
        qpd = self.qpd

        k_c = np.dot(qpd.gd.cell_cv, k_v) / (2 * np.pi)

        weight = np.sqrt(self.analyzer.get_kpoint_weight(k_c) /
                         self.analyzer.how_many_symmetries())

        # Here we're again setting pawcorr willy-nilly
        if self._chi0calc.pawcorr is None:
            pairden_paw_corr = self.gs.pair_density_paw_corrections
            self._chi0calc.pawcorr = pairden_paw_corr(qpd)

        kptpair = self.kptpair_factory.get_kpoint_pair(
            qpd, s, k_c, self.n1, self.n2,
            self.m1, self.m2, block=block)

        m_m = np.arange(self.m1, self.m2)
        n_n = np.arange(self.n1, self.n2)
        n_nmG = target_method(qpd, kptpair, n_n, m_m,
                              pawcorr=self._chi0calc.pawcorr,
                              block=block)

        if self.integrationmode is None:
            n_nmG *= weight

        df_nm = kptpair.get_occupation_differences(n_n, m_m)
        df_nm[df_nm <= 1e-20] = 0.0
        n_nmG *= df_nm[..., np.newaxis]**0.5

        return n_nmG

    @timer('Get eigenvalues')
    def eigenvalues(self, k_v, s):
        """A function that can return the eigenvalues.

        A simple function describing the integrand of
        the response function which gives an output that
        is compatible with the gpaw k-point integration
        routines."""

        qpd = self.qpd
        gs = self.gs
        kd = gs.kd

        k_c = np.dot(qpd.gd.cell_cv, k_v) / (2 * np.pi)
        K1 = self.kptpair_factory.find_kpoint(k_c)
        K2 = self.kptpair_factory.find_kpoint(k_c + qpd.q_c)

        ik1 = kd.bz2ibz_k[K1]
        ik2 = kd.bz2ibz_k[K2]
        kpt1 = gs.kpt_qs[ik1][s]
        assert kd.comm.size == 1
        kpt2 = gs.kpt_qs[ik2][s]
        deps_nm = np.subtract(kpt1.eps_n[self.n1:self.n2][:, np.newaxis],
                              kpt2.eps_n[self.m1:self.m2])
        return deps_nm.reshape(-1)


class Chi0Calculator:
    def __init__(self, *args,
                 eshift=0.0,
                 intraband=True,
                 rate=0.0,
                 **kwargs):
        self.chi0_body_calc = Chi0BodyCalculator(
            *args, eshift=eshift, **kwargs)
        self.chi0_opt_ext_calc = Chi0OpticalExtensionCalculator(
            *args, intraband=intraband, rate=rate, **kwargs)

        # Attributes groped by other classes...
        # Oh the horror, there are many of these...
        # Remove these in the future XXX
        self.gs = self.chi0_body_calc.gs
        self.context = self.chi0_body_calc.context
        self.wd = self.chi0_body_calc.wd
        self.nbands = self.chi0_body_calc.nbands
        self.nocc1 = self.chi0_body_calc.nocc1
        self.nocc2 = self.chi0_body_calc.nocc2
        self.ecut = self.chi0_body_calc.ecut
        self.kptpair_factory = self.chi0_body_calc.kptpair_factory
        self.integrator = self.chi0_body_calc.integrator

    @property
    def nblocks(self):
        return self.kptpair_factory.nblocks

    @property
    def pair_calc(self):
        return self.kptpair_factory.pair_calculator()

    def base_ini(self, kptpair_factory,
                 context=None,
                 disable_point_group=False,
                 disable_time_reversal=False,
                 integrationmode=None):
        """Set up attributes common to all response function calculators."""

        if context is None:
            context = kptpair_factory.context

        # TODO: More refactoring to avoid non-orthogonal inputs.
        assert kptpair_factory.context.comm is context.comm
        self.context = context

        self.kptpair_factory = kptpair_factory
        self.gs = kptpair_factory.gs

        # Number of completely filled bands and number of non-empty bands.
        self.nocc1, self.nocc2 = self.gs.count_occupied_bands()

        self.disable_point_group = disable_point_group
        self.disable_time_reversal = disable_time_reversal

        # Set up integrator
        self.integrationmode = integrationmode
        self.integrator = self.construct_integrator()

    def tmp_init(self, wd, *args,
                 hilbert=True,
                 nbands=None,
                 timeordered=False,
                 ecut=None,
                 eta=0.2,
                 **kwargs):
        """Set up attributes to calculate the chi0 body and optical extensions.
        """
        self.base_ini(*args, **kwargs)

        if ecut is None:
            ecut = 50.0
        ecut /= Ha
        self.ecut = ecut
        self.nbands = nbands or self.gs.bd.nbands

        self.wd = wd
        self.context.print(self.wd, flush=False)

        self.eta = eta / Ha
        self.hilbert = hilbert
        self.task = self.construct_integral_task()

        self.timeordered = bool(timeordered)
        if self.timeordered:
            assert self.hilbert  # Timeordered is only needed for G0W0

        self.pawcorr = None

        self.context.print('Nonperiodic BCs: ', (~self.pbc))
        if sum(self.pbc) == 1:
            raise ValueError('1-D not supported atm.')

    def construct_integral_task(self):
        if self.eta == 0:
            assert not self.hilbert
            # eta == 0 is used as a synonym for calculating the hermitian part
            # of the response function at a range of imaginary frequencies
            assert not self.wd.omega_w.real.any()
            return self.construct_hermitian_task()

        if self.hilbert:
            # The hilbert flag is used to calculate the reponse function via a
            # hilbert transform of its dissipative (spectral) part.
            assert isinstance(self.wd, NonLinearFrequencyDescriptor)
            return self.construct_hilbert_task()

        # Otherwise, we perform a literal evaluation of the response function
        # at the given frequencies with broadening eta
        return self.construct_literal_task()

    def construct_hilbert_task(self):
        if isinstance(self.integrator, PointIntegrator):
            return self.construct_point_hilbert_task()
        else:
            assert isinstance(self.integrator, TetrahedronIntegrator)
            return self.construct_tetra_hilbert_task()

    def construct_tetra_hilbert_task(self):
        return HilbertTetrahedron(self.integrator.blockcomm)

    @property
    def pbc(self):
        return self.gs.pbc

    def create_chi0(self, q_c):
        chi0_body = self.chi0_body_calc.create_chi0_body(q_c)
        chi0 = Chi0Data.from_chi0_body_data(chi0_body)
        return chi0

    def get_pw_descriptor(self, q_c):
        return SingleQPWDescriptor.from_q(q_c, self.ecut, self.gs.gd)

    def calculate(self, q_c, spin='all'):
        """Calculate chi0 (possibly with optical extensions).

        Parameters
        ----------
        q_c : list or ndarray
            Momentum vector.
        spin : str or int
            If 'all' then include all spins.
            If 0 or 1, only include this specific spin.

        Returns
        -------
        chi0 : Chi0Data
            Data object containing the chi0 data arrays along with basis
            representation descriptors and blocks distribution
        """
        # Calculate body
        chi0_body = self.chi0_body_calc.calculate(q_c, spin=spin)
        qpd = chi0_body.qpd

        # Calculate optical extension
        if qpd.optical_limit:
            if not abs(self.chi0_body_calc.eshift) < 1e-8:
                raise NotImplementedError("No wings eshift available")
            chi0_opt_ext = self.chi0_opt_ext_calc.calculate(qpd=qpd, spin=spin)
        else:
            chi0_opt_ext = None

        self.context.print('\nFinished calculating chi0\n')

        return Chi0Data(chi0_body, chi0_opt_ext)

    def get_band_transitions(self):
        return self.nocc1, self.nbands  # m1, m2

    def get_spins(self, spin='all'):
        nspins = self.gs.nspins
        if spin == 'all':
            spins = range(nspins)
        else:
            assert spin in range(nspins)
            spins = [spin]

        return spins

    @timer('Calculate CHI_0')
    def update_chi0(self,
                    chi0: Chi0Data,
                    m1, m2, spins):
        """In-place calculation of chi0 (with optical extension).

        Parameters
        ----------
        chi0 : Chi0Data
            Data and representation object
        m1 : int
            Lower band cutoff for band summation
        m2 : int
            Upper band cutoff for band summation
        spins : list
            List of spin indices to include in the calculation

        Returns
        -------
        chi0 : Chi0Data
        """
        self.chi0_body_calc.update_chi0_body(chi0.body, m1, m2, spins)
        if chi0.optical_limit:
            if not abs(self.chi0_body_calc.eshift) < 1e-8:
                raise NotImplementedError("No wings eshift available")
            assert chi0.optical_extension is not None
            # Update the head and wings
            self.chi0_opt_ext_calc.update_chi0_optical_extension(
                chi0.optical_extension, m1, m2, spins)
        return chi0

    def construct_integrator(self):
        """Construct k-point integrator"""
        cls = self.get_integrator_cls()
        return cls(
            cell_cv=self.gs.gd.cell_cv,
            context=self.context,
            nblocks=self.nblocks)

    def get_integrator_cls(self):
        """Get the appointed k-point integrator class."""
        if self.integrationmode is None:
            self.context.print('Using integrator: PointIntegrator')
            cls = PointIntegrator
        elif self.integrationmode == 'tetrahedron integration':
            self.context.print('Using integrator: TetrahedronIntegrator')
            cls = TetrahedronIntegrator  # type: ignore
            if not all([self.disable_point_group,
                        self.disable_time_reversal]):
                self.check_high_symmetry_ibz_kpts()
        else:
            raise ValueError(f'Integration mode "{self.integrationmode}"'
                             ' not implemented.')
        return cls

    def check_high_symmetry_ibz_kpts(self):
        """Check that the ground state includes all corners of the IBZ."""
        ibz_vertices_kc = self.gs.get_ibz_vertices()
        # Here we mimic the k-point grid compatibility check of
        # gpaw.bztools.find_high_symmetry_monkhorst_pack()
        bzk_kc = self.gs.kd.bzk_kc
        for ibz_vertex_c in ibz_vertices_kc:
            # Relative coordinate difference to the k-point grid
            diff_kc = np.abs(bzk_kc - ibz_vertex_c)[:, self.gs.pbc].round(6)
            # The ibz vertex should exits in the BZ grid up to a reciprocal
            # lattice vector, meaning that the relative coordinate difference
            # is allowed to be an integer. Thus, at least one relative k-point
            # difference should vanish, modulo 1
            mod_diff_kc = np.mod(diff_kc, 1)
            nodiff_k = np.all(mod_diff_kc < 1e-5, axis=1)
            if not np.any(nodiff_k):
                raise ValueError(
                    'The ground state k-point grid does not include all '
                    'vertices of the IBZ. '
                    'Please use find_high_symmetry_monkhorst_pack() from '
                    'gpaw.bztools to generate your k-point grid.')

    def get_integration_domain(self, qpd, spins):
        """Get integrator domain and prefactor for the integral."""
        for spin in spins:
            assert spin in range(self.gs.nspins)
        # The integration domain is determined by the following function
        # that reduces the integration domain to the irreducible zone
        # of the little group of q.
        bzk_kv, analyzer = self.get_kpoints(
            qpd, integrationmode=self.integrationmode)
        domain = (bzk_kv, spins)

        if self.integrationmode == 'tetrahedron integration':
            # If there are non-periodic directions it is possible that the
            # integration domain is not compatible with the symmetry operations
            # which essentially means that too large domains will be
            # integrated. We normalize by vol(BZ) / vol(domain) to make
            # sure that to fix this.
            domainvol = convex_hull_volume(
                bzk_kv) * analyzer.how_many_symmetries()
            bzvol = (2 * np.pi)**3 / self.gs.volume
            factor = bzvol / domainvol
        else:
            factor = 1

        prefactor = (2 * factor * analyzer.how_many_symmetries() /
                     (self.gs.nspins * (2 * np.pi)**3))  # Remember prefactor

        if self.integrationmode is None:
            nbzkpts = self.gs.kd.nbzkpts
            prefactor *= len(bzk_kv) / nbzkpts

        return domain, analyzer, prefactor

    @timer('Get kpoints')
    def get_kpoints(self, qpd, integrationmode):
        """Get the integration domain."""
        analyzer = PWSymmetryAnalyzer(
            self.gs.kd, qpd, self.context,
            disable_point_group=self.disable_point_group,
            disable_time_reversal=self.disable_time_reversal)

        if integrationmode is None:
            K_gK = analyzer.group_kpoints()
            bzk_kc = np.array([self.gs.kd.bzk_kc[K_K[0]] for
                               K_K in K_gK])
        elif integrationmode == 'tetrahedron integration':
            bzk_kc = analyzer.get_reduced_kd(pbc_c=self.pbc).bzk_kc
            if (~self.pbc).any():
                bzk_kc = np.append(bzk_kc,
                                   bzk_kc + (~self.pbc).astype(int),
                                   axis=0)

        bzk_kv = np.dot(bzk_kc, qpd.gd.icell_cv) * 2 * np.pi
        return bzk_kv, analyzer

    def get_gs_info_string(self, tab=''):
        gs = self.gs
        gd = gs.gd

        ns = gs.nspins
        nk = gs.kd.nbzkpts
        nik = gs.kd.nibzkpts

        nocc = self.nocc1
        npocc = self.nocc2
        ngridpoints = gd.N_c[0] * gd.N_c[1] * gd.N_c[2]
        nstat = ns * npocc
        occsize = nstat * ngridpoints * 16. / 1024**2

        gs_list = [f'{tab}Ground state adapter containing:',
                   f'Number of spins: {ns}', f'Number of kpoints: {nk}',
                   f'Number of irreducible kpoints: {nik}',
                   f'Number of completely occupied states: {nocc}',
                   f'Number of partially occupied states: {npocc}',
                   f'Occupied states memory: {occsize} M / cpu']

        return f'\n{tab}'.join(gs_list)

    def get_response_info_string(self, qpd, tab=''):
        nw = len(self.wd)
        ecut = self.ecut * Ha
        nbands = self.nbands
        ngmax = qpd.ngmax
        eta = self.eta * Ha

        res_list = [f'{tab}Number of frequency points: {nw}',
                    f'Planewave cutoff: {ecut}',
                    f'Number of bands: {nbands}',
                    f'Number of planewaves: {ngmax}',
                    f'Broadening (eta): {eta}']

        return f'\n{tab}'.join(res_list)


class Chi0BodyCalculator(Chi0Calculator):

    def __init__(self, *args,
                 eshift=0.0,
                 **kwargs):
        self.eshift = eshift / Ha
        self.tmp_init(*args, **kwargs)

        metallic = self.nocc1 != self.nocc2
        if metallic:
            assert abs(self.eshift) < 1e-8, \
                'A rigid energy shift cannot be applied to the conduction '\
                'bands if there is no band gap'

    def create_chi0_body(self, q_c):
        qpd = self.get_pw_descriptor(q_c)
        return self._create_chi0_body(qpd)

    def _create_chi0_body(self, qpd):
        return Chi0BodyData(self.wd, qpd, self.get_blockdist())

    def get_blockdist(self):
        return PlaneWaveBlockDistributor(
            self.context.comm,
            self.integrator.blockcomm,
            self.integrator.kncomm)

    def calculate(self, q_c, spin='all') -> Chi0BodyData:
        """Calculate the chi0 body.

        Parameters
        ----------
        q_c : list or ndarray
            Momentum vector.
        spin : str or int
            If 'all' then include all spins.
            If 0 or 1, only include this specific spin."""
        # Construct the output data structure
        qpd = self.get_pw_descriptor(q_c)
        self.print_info(qpd)
        chi0_body = self._create_chi0_body(qpd)

        # Integrate all transitions into partially filled and empty bands
        m1, m2 = self.get_band_transitions()
        spins = self.get_spins(spin)
        self.update_chi0_body(chi0_body, m1, m2, spins)

        return chi0_body

    def update_chi0_body(self,
                         chi0_body: Chi0BodyData,
                         m1, m2, spins):
        """In-place calculation of the body."""
        qpd = chi0_body.qpd

        # Reset PAW correction in case momentum has change
        pairden_paw_corr = self.gs.pair_density_paw_corrections
        self.pawcorr = pairden_paw_corr(chi0_body.qpd)

        self.context.print('Integrating chi0 body.')

        domain, analyzer, prefactor = self.get_integration_domain(qpd, spins)
        integrand = Chi0Integrand(self, qpd=qpd, analyzer=analyzer,
                                  optical=False, m1=m1, m2=m2)

        chi0_body.data_WgG[:] /= prefactor
        if self.hilbert:
            # Allocate a temporary array for the spectral function
            out_WgG = chi0_body.zeros()
        else:
            # Use the preallocated array for direct updates
            out_WgG = chi0_body.data_WgG
        self.integrator.integrate(domain=domain,  # Integration domain
                                  integrand=integrand,
                                  task=self.task,
                                  wd=self.wd,  # Frequency Descriptor
                                  out_wxx=out_WgG)  # Output array

        if self.hilbert:
            # The integrator only returns the spectral function and a Hilbert
            # transform is performed to return the real part of the density
            # response function.
            with self.context.timer('Hilbert transform'):
                # Make Hilbert transform
                ht = HilbertTransform(np.array(self.wd.omega_w), self.eta,
                                      timeordered=self.timeordered)
                ht(out_WgG)
            # Update the actual chi0 array
            chi0_body.data_WgG[:] += out_WgG
        chi0_body.data_WgG[:] *= prefactor

        tmp_chi0_wGG = chi0_body.copy_array_with_distribution('wGG')
        analyzer.symmetrize_wGG(tmp_chi0_wGG)
        chi0_body.data_WgG[:] = chi0_body.blockdist.distribute_as(
            tmp_chi0_wGG, chi0_body.nw, 'WgG')

    def construct_hermitian_task(self):
        return Hermitian(self.integrator.blockcomm, eshift=self.eshift)

    def construct_point_hilbert_task(self):
        return Hilbert(self.integrator.blockcomm, eshift=self.eshift)

    def construct_literal_task(self):
        return GenericUpdate(
            self.eta, self.integrator.blockcomm, eshift=self.eshift)

    def print_info(self, qpd):

        if gpaw.dry_run:
            from gpaw.mpi import SerialCommunicator
            size = gpaw.dry_run
            comm = SerialCommunicator()
            comm.size = size
        else:
            comm = self.context.comm

        q_c = qpd.q_c
        nw = len(self.wd)
        csize = comm.size
        knsize = self.integrator.kncomm.size
        bsize = self.integrator.blockcomm.size
        chisize = nw * qpd.ngmax**2 * 16. / 1024**2 / bsize

        isl = ['', f'{ctime()}',
               'Calculating chi0 body with:',
               self.get_gs_info_string(tab='    '), '',
               '    Linear response parametrization:',
               f'    q_c: [{q_c[0]}, {q_c[1]}, {q_c[2]}]',
               self.get_response_info_string(qpd, tab='    '),
               f'    comm.size: {csize}',
               f'    kncomm.size: {knsize}',
               f'    blockcomm.size: {bsize}']
        if bsize > nw:
            isl.append(
                'WARNING! Your nblocks is larger than number of frequency'
                ' points. Errors might occur, if your submodule does'
                ' not know how to handle this.')
        isl.extend(['',
                    '    Memory estimate of potentially large arrays:',
                    f'        chi0_wGG: {chisize} M / cpu',
                    '        Memory usage before allocation: '
                    f'{(maxrss() / 1024**2)} M / cpu'])
        self.context.print('\n'.join(isl))


class Chi0OpticalExtensionCalculator(Chi0Calculator):

    def __init__(self, *args,
                 intraband=True,
                 rate=0.0,
                 **kwargs):
        self.tmp_init(*args, **kwargs)

        # In the optical limit of metals, one must add the Drude dielectric
        # response from the free-space plasma frequency of the intraband
        # transitions to the head of the chi0 wings. This is handled by a
        # separate calculator, provided that intraband is set to True.
        metallic = self.nocc1 != self.nocc2
        if metallic and intraband:
            from gpaw.response.drude import Chi0DrudeCalculator
            if rate == 'eta':
                rate = self.eta * Ha  # external units
            self.rate = rate
            self.drude_calc = Chi0DrudeCalculator(
                self.kptpair_factory,
                disable_point_group=self.disable_point_group,
                disable_time_reversal=self.disable_time_reversal,
                integrationmode=self.integrationmode)
        else:
            self.drude_calc = None
            self.rate = None

    @property
    def nblocks(self):
        # The optical extensions are not distributed in memory
        # NB: There can be a mismatch with
        # self.kptpair_factory.nblocks, which seems dangerous XXX
        return 1

    def calculate(self,
                  qpd: SingleQPWDescriptor | None = None,
                  spin='all'):
        """Calculate the chi0 head and wings.

        Paramters
        ---------
        spin : str or int
            If 'all' then include all spins.
            If 0 or 1, only include this specific spin.
        """
        # Create data object
        if qpd is None:
            qpd = self.get_pw_descriptor(q_c=[0., 0., 0.])
        chi0_opt_ext = Chi0OpticalExtensionData(self.wd, qpd)

        self.print_info(qpd)

        # Define band and spin transitions
        m1, m2 = self.get_band_transitions()
        spins = self.get_spins(spin)

        # Perform the actual integration
        self.update_chi0_optical_extension(chi0_opt_ext, m1, m2, spins)

        if self.drude_calc is not None:
            # Add intraband contribution
            chi0_drude = self.drude_calc.calculate(self.wd, self.rate, spin)
            chi0_opt_ext.head_Wvv[:] += chi0_drude.chi_Zvv

        return chi0_opt_ext

    def update_chi0_optical_extension(
            self,
            chi0_optical_extension: Chi0OpticalExtensionData,
            m1, m2, spins):
        """In-place calculation of the chi0 head and wings.

        Parameters
        ----------
        m1 : int
            Lower band cutoff for band summation
        m2 : int
            Upper band cutoff for band summation
        spins : list
            List of spin indices to include in the calculation
        """
        self.context.print('Integrating chi0 head and wings.')
        chi0_opt_ext = chi0_optical_extension
        qpd = chi0_opt_ext.qpd

        domain, analyzer, prefactor = self.get_integration_domain(qpd, spins)
        integrand = Chi0Integrand(self, qpd=qpd, analyzer=analyzer,
                                  optical=True, m1=m1, m2=m2)

        # We integrate the head and wings together, using the combined index P
        # index v = (x, y, z)
        # index G = (G0, G1, G2, ...)
        # index P = (x, y, z, G1, G2, ...)
        WxvP_shape = list(chi0_opt_ext.WxvG_shape)
        WxvP_shape[-1] += 2
        tmp_chi0_WxvP = np.zeros(WxvP_shape, complex)
        self.integrator.integrate(domain=domain,  # Integration domain
                                  integrand=integrand,
                                  task=self.task,
                                  wd=self.wd,  # Frequency Descriptor
                                  out_wxx=tmp_chi0_WxvP)  # Output array
        if self.hilbert:
            with self.context.timer('Hilbert transform'):
                ht = HilbertTransform(np.array(self.wd.omega_w), self.eta,
                                      timeordered=self.timeordered)
                ht(tmp_chi0_WxvP)
        tmp_chi0_WxvP *= prefactor

        # Fill in wings part of the data, but leave out the head part (G0)
        chi0_opt_ext.wings_WxvG[..., 1:] += tmp_chi0_WxvP[..., 3:]
        analyzer.symmetrize_wxvG(chi0_opt_ext.wings_WxvG)
        # Fill in the head
        chi0_opt_ext.head_Wvv[:] += tmp_chi0_WxvP[:, 0, :3, :3]
        analyzer.symmetrize_wvv(chi0_opt_ext.head_Wvv)

    def construct_hermitian_task(self):
        return HermitianOpticalLimit()

    def construct_point_hilbert_task(self):
        return HilbertOpticalLimit()

    def construct_tetra_hilbert_task(self):
        return HilbertOpticalLimitTetrahedron()

    def construct_literal_task(self):
        return OpticalLimit(eta=self.eta)

    def print_info(self, qpd):
        """Print information about optical extension calculation."""
        isl = ['',
               f'{ctime()}',
               'Calculating chi0 optical extensions with:',
               self.get_gs_info_string(tab='    '),
               '',
               '    Linear response parametrization:',
               self.get_response_info_string(qpd, tab='    ')]
        self.context.print('\n'.join(isl))


class Chi0(Chi0Calculator):
    """Class for calculating non-interacting response functions.
    Tries to be backwards compatible, for now. """

    def __init__(self,
                 calc,
                 *,
                 frequencies: Union[dict, Array1D] = None,
                 ecut=50,
                 world=mpi.world, txt='-', timer=None,
                 nblocks=1,
                 nbands=None,
                 domega0=None,  # deprecated
                 omega2=None,  # deprecated
                 omegamax=None,  # deprecated
                 **kwargs):
        """Construct Chi0 object.

        Parameters
        ----------
        calc : str
            The groundstate calculation file that the linear response
            calculation is based on.
        frequencies :
            Input parameters for frequency_grid.
            Can be array of frequencies to evaluate the response function at
            or dictionary of paramaters for build-in nonlinear grid
            (see :ref:`frequency grid`).
        ecut : float
            Energy cutoff.
        hilbert : bool
            Switch for hilbert transform. If True, the full density response
            is determined from a hilbert transform of its spectral function.
            This is typically much faster, but does not work for imaginary
            frequencies.
        nbands : int
            Maximum band index to include.
        timeordered : bool
            Switch for calculating the time ordered density response function.
            In this case the hilbert transform cannot be used.
        eta : float
            Artificial broadening of spectra.
        intraband : bool
            Switch for including the intraband contribution to the density
            response function.
        world : MPI comm instance
            MPI communicator.
        txt : str
            Output file.
        timer : gpaw.utilities.timing.timer instance
        nblocks : int
            Divide the response function into nblocks. Useful when the response
            function is large.
        disable_point_group : bool
            Do not use the point group symmetry operators.
        disable_time_reversal : bool
            Do not use time reversal symmetry.
        integrationmode : str
            Integrator for the kpoint integration.
            If == 'tetrahedron integration' then the kpoint integral is
            performed using the linear tetrahedron method.
        eshift : float
            Shift unoccupied bands
        rate : float,str
            Phenomenological scattering rate to use in optical limit Drude term
            (in eV). If rate='eta', then use input artificial broadening eta as
            rate. Note, for consistency with the formalism the rate is
            implemented as omegap^2 / (omega + 1j * rate)^2 which differ from
            some literature by a factor of 2.


        Attributes
        ----------
        kptpair_factory : gpaw.response.pair.KPointPairFactory instance
            Class for calculating matrix elements of pairs of wavefunctions.

        """
        from gpaw.response.pair import get_gs_and_context
        gs, context = get_gs_and_context(calc, txt, world, timer)
        nbands = nbands or gs.bd.nbands

        wd = new_frequency_descriptor(
            gs, context, nbands, frequencies,
            domega0=domega0,
            omega2=omega2, omegamax=omegamax)

        kptpair_factory = KPointPairFactory(gs, context, nblocks=nblocks)

        super().__init__(wd=wd, kptpair_factory=kptpair_factory,
                         nbands=nbands, ecut=ecut, **kwargs)


def new_frequency_descriptor(gs, context, nbands, frequencies=None, *,
                             domega0=None, omega2=None, omegamax=None):
    if domega0 is not None or omega2 is not None or omegamax is not None:
        assert frequencies is None
        frequencies = {'type': 'nonlinear',
                       'domega0': domega0,
                       'omega2': omega2,
                       'omegamax': omegamax}
        warnings.warn(f'Please use frequencies={frequencies}')

    elif frequencies is None:
        frequencies = {'type': 'nonlinear'}

    if (isinstance(frequencies, dict) and
        frequencies.get('omegamax') is None):
        omegamax = find_maximum_frequency(gs.kpt_u, context,
                                          nbands=nbands)
        frequencies['omegamax'] = omegamax * Ha

    wd = FrequencyDescriptor.from_array_or_dict(frequencies)
    return wd



import pytest
import numpy as np
from gpaw.response.g0w0 import G0W0


@pytest.mark.response
@pytest.mark.parametrize('fxc_mode, ref_gap', [
    ('GWP', 4.667170),
    ('GWS', 4.988230),
    ('GWG', 4.894904)])
def test_fxc_mode(in_tmp_dir, gpw_files, fxc_mode, ref_gap, scalapack,
                  needs_ase_master):
    gw = G0W0(gpw_files['bn_pw'],
              bands=(3, 5),
              nbands=9,
              nblocks=1,
              xc='rALDA',
              ecut=40,
              fxc_mode=fxc_mode)

    result = gw.calculate()

    calculated_gap = np.min(result['qp'][0, :, 1])\
        - np.max(result['qp'][0, :, 0])
    assert calculated_gap == pytest.approx(ref_gap, abs=1e-4)



import pytest
from gpaw import GPAW


def test_atomic_el_pot(gpw_files):
    calc = GPAW(gpw_files['h2_pw'])
    values = calc.get_atomic_electrostatic_potentials()
    ref = -49.48543067
    assert values == pytest.approx([ref, ref])



import pytest
import numpy as np

from gpaw.basis_data import Basis
from gpaw.mpi import world
from gpaw.lcao.generate_ngto_augmented import \
    create_GTO_dictionary as GTO, create_CGTO_dictionary as CGTO, \
    generate_nao_ngto_basis, read_gaussian_basis_file

pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


def test_read_gaussian(in_tmp_dir):
    atom = 'Au'
    description = 'test description\nline2'
    gtos = [GTO(0, 0.2),
            GTO('s', 0.1),
            GTO(1, 0.05),
            CGTO(1, [9.2, 2.1, 0.54, 0.15], [0.02, 0.18, 0.3, 0.5]),
            CGTO(2, [0.05], [1.0])]

    with open('gbs.txt', 'w') as f:
        def w(s):
            f.write('%s\n' % s)
        w('****')
        for desc in description.split('\n'):
            w(f'! {desc}')
        w(f'{atom} 0')
        w('S   1   1.00')
        w('      0.2000000              1.000000D+00')
        w('S   1   1.00')
        w('      0.1000000              1.0000000')
        w('P   1   1.00')
        w('      0.0500000              1.0000000')
        w('P   4   1.00')
        w('      9.200000D+00           2.000000E-02')
        w('      2.100000E+00           1.800000D-01')
        w('      5.400000e-01           3.000000D-01')
        w('      1.500000D-01           5.000000e-01')
        w('D   1   1.00')
        w('      0.500000D-01           1.0000000')
        w('****')
        w('This line is never read')

    gbs_atom, gbs_description, gbs_gtos = read_gaussian_basis_file('gbs.txt')
    assert gbs_atom == atom
    assert gbs_description == description
    assert len(gbs_gtos) == len(gtos)
    for gbs_gto, gto in zip(gbs_gtos, gtos):
        for key in gto:
            assert np.allclose(gbs_gto[key], gto[key])


def test_generate(in_tmp_dir):
    with open('gbs.txt', 'w') as f:
        def w(s):
            f.write('%s\n' % s)
        w('C     0')
        w('S   1   1.00')
        w('      1.596000D-01           1.000000D+00')
        w('S   1   1.00')
        w('      0.0469000              1.0000000')
        w('P   4   1.00')
        w('      9.439000D+00           3.810900D-02')
        w('      2.002000D+00           2.094800D-01')
        w('      5.456000D-01           5.085570D-01')
        w('      1.517000D-01           4.688420D-01')
        w('P   1   1.00')
        w('      1.517000D-01           1.000000D+00')
        w('P   1   1.00')
        w('      0.0404100              1.0000000')
        w('D   1   1.00')
        w('      5.500000D-01           1.0000000')
        w('D   1   1.00')
        w('      0.1510000              1.0000000')

    gbs_atom, gbs_description, gtos = read_gaussian_basis_file('gbs.txt')

    assert gbs_atom == 'C'
    generate_nao_ngto_basis('C', xc='LDA', nao='dzp', name='NAO+NGTO',
                            gtos=gtos, gto_description=gbs_description)

    basis = Basis('C', 'NAO+NGTO', readxml=False)
    basis.read_xml('C.NAO+NGTO.dzp.basis')
    assert len(basis.bf_j) == 5 + 7



import numpy as np

from ase.units import Bohr

from gpaw import debug
from gpaw.tddft import aufrequency_to_eV
from gpaw.transformers import Transformer
from gpaw.lcaotddft.observer import TDDFTObserver
from gpaw.utilities import is_contiguous

from gpaw.inducedfield.inducedfield_base import BaseInducedField


class FDTDInducedField(BaseInducedField, TDDFTObserver):
    """Induced field class for FDTD.

    Attributes (see also ``BaseInducedField``):
    -------------------------------------------
    time: float
        Current time
    Fn_wG: ndarray (complex)
        Fourier transform of induced polarization charge density
    n0_G: ndarray (float)
        Ground state charge density
    """

    def __init__(self, filename=None, paw=None,
                 frequencies=None, folding='Gauss', width=0.08,
                 interval=1, restart_file=None):
        """
        Parameters (see also ``BaseInducedField``):
        -------------------------------------------
        paw: TDDFT object
            TDDFT object for time propagation
        width: float
            Width in eV for the Gaussian (sigma) or Lorentzian (eta) folding
            Gaussian   = exp(- (1/2) * sigma^2 * t^2)
            Lorentzian = exp(- eta * t)
        interval: int
            Number of timesteps between calls (used when attaching)
        restart_file: string
            Name of the restart file
        """

        TDDFTObserver.__init__(self, paw, interval)
        # From observer:
        # self.niter
        # self.interval
        # self.timer
        # Observer does also paw.attach(self, ...)

        # Restart file
        self.restart_file = restart_file

        # These are allocated in allocate()
        self.Fn_wsG = None
        self.n0_G = None

        self.readwritemode_str_to_list = {
            '': ['Fn', 'n0', 'atoms'],
            'all': ['Fn', 'n0',
                    'Frho', 'Fphi', 'Fef', 'Ffe', 'eps0', 'atoms'],
            'field': ['Frho', 'Fphi', 'Fef', 'Ffe', 'eps0', 'atoms']}

        BaseInducedField.__init__(self, filename, paw,
                                  frequencies, folding, width)

    def initialize(self, paw, allocate=True):
        BaseInducedField.initialize(self, paw, allocate)

        if self.has_paw:
            # FDTD replacements and overwrites
            self.fdtd = paw.hamiltonian.poisson
            # self.gd = self.fdtd.cl.gd.refine()
            self.gd = self.fdtd.cl.gd

            assert hasattr(paw, 'time') and hasattr(paw, 'niter'), 'Use TDDFT!'
            self.time = paw.time
            self.niter = paw.niter

    def set_folding(self, folding, width):
        BaseInducedField.set_folding(self, folding, width)

        if self.folding is None:
            self.envelope = lambda t: 1.0
        else:
            if self.folding == 'Gauss':
                self.envelope = lambda t: np.exp(- 0.5 * self.width**2 * t**2)
            elif self.folding == 'Lorentz':
                self.envelope = lambda t: np.exp(- self.width * t)
            else:
                raise RuntimeError('unknown folding "' + self.folding + '"')

    def allocate(self):
        if not self.allocated:

            gd = self.paw.hamiltonian.poisson.cl.gd
            # Ground state charge density
            self.n0_G = gd.empty() + np.nan

            # Fourier transformed charge density
            self.Fn_wG = gd.zeros((self.nw,), dtype=self.dtype)
            self.allocated = True

        if debug:
            assert is_contiguous(self.Fn_wG, self.dtype)

    def deallocate(self):
        BaseInducedField.deallocate(self)
        self.n0_G = None
        self.Fn_wG = None

    def _update(self, paw):
        if paw.action == 'init':
            if paw.niter == 0:
                c = self.paw.hamiltonian.poisson.classical_material
                self.n0_G[:] = -1 * c.sign * c.charge_density
            return
        elif paw.action == 'kick':
            # Background electric field
            self.Fbgef_v = paw.kick_strength
            return
        elif paw.action != 'propagate':
            return

        assert (self.Fbgef_v is not None
                and not np.any(np.isnan(self.n0_G))), \
            f'Attach {self.__class__.__name__} before absorption kick'

        # Update time
        self.time = paw.time
        time_step = paw.time_step

        # Complex exponential with envelope
        f_w = (np.exp(1.0j * self.omega_w * self.time) *
               self.envelope(self.time) * time_step)

        # Time-dependent quantities
        n_G = (-1.0 * self.fdtd.classical_material.sign *
               self.fdtd.classical_material.charge_density)

        # Update Fourier transforms
        for w in range(self.nw):
            self.Fn_wG[w] += (n_G - self.n0_G) * f_w[w]

        # Restart file
        # XXX remove this once deprecated dump_interval is removed,
        # but keep write_restart() as it'll be still used
        # (see TDDFTObserver class)
        if (paw.restart_file is not None
                and self.niter % paw.dump_interval == 0):
            self.write_restart()

    def write_restart(self):
        if self.restart_file is not None:
            self.write(self.restart_file)
            self.log(f'{self.__class__.__name__}: Wrote restart file')

    def get_induced_density(self, from_density, gridrefinement):
        if self.gd == self.fdtd.cl.gd:
            Frho_wg = self.Fn_wG.copy()
        else:
            Frho_wg = Transformer(self.fdtd.cl.gd,
                                  self.gd,
                                  self.stencil,
                                  dtype=self.dtype).apply(self.Fn_wG)

        Frho_wg, gd = self.interpolate_density(self.gd, Frho_wg,
                                               gridrefinement)
        return Frho_wg, gd

    def interpolate_density(self, gd, Fn_wg, gridrefinement=2):

        # Find m for
        # gridrefinement = 2**m
        m1 = np.log(gridrefinement) / np.log(2.)
        m = int(np.round(m1))

        # Check if m is really integer
        if np.absolute(m - m1) < 1e-8:
            for i in range(m):
                gd2 = gd.refine()

                # Interpolate
                interpolator = Transformer(gd, gd2, self.stencil,
                                           dtype=self.dtype)
                Fn2_wg = gd2.empty((self.nw,), dtype=self.dtype)
                for w in range(self.nw):
                    interpolator.apply(Fn_wg[w], Fn2_wg[w],
                                       np.ones((3, 2), dtype=complex))

                gd = gd2
                Fn_wg = Fn2_wg
        else:
            raise NotImplementedError

        return Fn_wg, gd

    def _read(self, reader, reads):
        BaseInducedField._read(self, reader, reads)

        # Test time
        r = reader
        time = r.time
        if self.has_paw:
            # Test time
            if abs(time - self.time) >= 1e-9:
                raise IOError('Timestamp is incompatible with calculator.')
        else:
            self.time = time

        # Allocate
        self.allocate()

        def readarray(name):
            if name.split('_')[0] in reads:
                self.gd.distribute(r.get(name), getattr(self, name))

        # Read arrays
        readarray('n0_G')
        readarray('Fn_wG')
        readarray('eps0_G')

    def _write(self, writer, writes):
        # Swap classical and quantum cells, and shift atom positions for
        # the time of writing
        qmcell = self.atoms.get_cell()
        self.atoms.set_cell(self.fdtd.cl.cell * Bohr)  # Set classical cell
        self.atoms.positions = (self.atoms.get_positions() +
                                self.fdtd.qm.corner1 * Bohr)
        BaseInducedField._write(self, writer, writes)
        self.atoms.set_cell(qmcell)  # Restore quantum cell to the atoms object
        self.atoms.positions = (self.atoms.get_positions() -
                                self.fdtd.qm.corner1 * Bohr)

        # Write time propagation status
        writer.write(time=self.time)

        # Mask, interpolation approach:
        # self.eps0_G = self.fdtd.classical_material.permittivityValue(
        # omega=0.0) - self.fdtd.classical_material.epsInfty
        # self.eps0_G= -interpolator.apply(self.eps0_G)

        # Mask, direct approach:
        self.eps0_G = self.fdtd.cl.gd.zeros()
        for component in self.fdtd.classical_material.components:
            self.eps0_G += 1.0 * component.get_mask(gd=self.fdtd.cl.gd)

#        def writearray(name, shape, dtype):
#            if name.split('_')[0] in writes:
#                writer.add_array(name, shape, dtype)
#            a_wxg = getattr(self, name)
#            for w in range(self.nw):
#                if self.fdtd.cl.gd == self.gd:
#                    writer.fill(self.gd.collect(a_wxg[w]))
#                else:
#                    writer.fill(self.gd.collect(Transformer(self.fdtd.cl.gd,
# self.gd, self.stencil, self.dtype).apply(a_wxg[w])))

        def writearray(name):
            if name.split('_')[0] in writes:
                a_xg = getattr(self, name)
                if self.fdtd.cl.gd != self.gd:
                    a_xg = Transformer(self.fdtd.cl.gd,
                                       self.gd,
                                       self.stencil,
                                       a_xg.dtype).apply(a_xg)
                writer.write(**{name: self.gd.collect(a_xg)})

        # Write time propagation arrays
        writearray('n0_G')
        writearray('Fn_wG')
        writearray('eps0_G')

        if hasattr(self.fdtd, 'qm') and hasattr(self.fdtd.qm, 'corner1'):
            writer.write(corner1_v=self.fdtd.qm.corner1)
            writer.write(corner2_v=self.fdtd.qm.corner2)

        self.fdtd.cl.gd.comm.barrier()


def calculate_hybrid_induced_field(cl_ind, qm_ind, h):
    """
    Combines two InducedField objects and
    calculates total induced field.

    Parameters:
    -----------
    cl_ind: FDTDInducedField
        InducedField object of the classical subsystem
    qm_ind: TDDFTInducedField
        InducedField object of the quantum subsystem
    h: float
        Grid spacing of total grid in Angstroms
    """
    # Sanity checks
    for attr in ['paw', 'omega_w', 'width', 'folding',
                 'Fbgef_v', 'world', 'atoms']:
        assert np.all(getattr(cl_ind, attr) == getattr(qm_ind, attr))

    # Object for the total induced field
    tot_ind = FDTDInducedField(paw=cl_ind.paw,
                               frequencies=cl_ind.omega_w * aufrequency_to_eV,
                               width=cl_ind.width * aufrequency_to_eV,
                               folding=cl_ind.folding)

    tot_ind.world = cl_ind.world
    tot_ind.Fbgef_v = cl_ind.Fbgef_v
    tot_ind.atoms = cl_ind.atoms.copy()
    tot_ind.na = cl_ind.na
    tot_ind.nspins = 1

    from_density = 'comp'
    cl_Frho_wg, cl_gd = cl_ind.get_induced_density(gridrefinement=1,
                                                   from_density=from_density)
    qm_Frho_wg, qm_gd = qm_ind.get_induced_density(gridrefinement=2,
                                                   from_density=from_density)

    fdtd_poisson = tot_ind.paw.hamiltonian.poisson

    def combine(qm_g, cl_g):
        return fdtd_poisson.get_combined_data(qm_g.copy(),
                                              cl_g.copy(),
                                              h,
                                              qm_gd,
                                              cl_gd)

    _, tot_gd = combine(qm_Frho_wg[0].imag, cl_Frho_wg[0].imag)
    tot_ind.gd = cl_ind.gd
    tot_ind.fieldgd = tot_gd
    tot_ind.Frho_wg = tot_gd.empty((tot_ind.nw,), dtype=complex)
    for w in range(tot_ind.nw):
        tot_Frho_g, _ = combine(qm_Frho_wg[w].real, cl_Frho_wg[w].real)
        tot_ind.Frho_wg[w][:] = tot_Frho_g
        tot_Frho_g, _ = combine(qm_Frho_wg[w].imag, cl_Frho_wg[w].imag)
        tot_ind.Frho_wg[w] += 1.0j * tot_Frho_g

    tot_ind.has_field = True
    tot_ind.field_from_density = from_density
    tot_ind.calculate_induced_field(from_density=from_density)
    return tot_ind



from math import sqrt
import numpy as np

from gpaw.symmetry import Symmetry
from ase.dft.kpoints import monkhorst_pack

# Primitive diamond lattice, with Si lattice parameter


def test_symmetry_symmetry_ft():
    a = 5.475
    cell_cv = .5 * a * np.array([(1, 1, 0), (1, 0, 1), (0, 1, 1)])
    spos_ac = np.array([(.00, .00, .00),
                        (.25, .25, .25)])
    id_a = [1, 1]  # Two identical atoms
    pbc_c = np.ones(3, bool)
    bzk_kc = monkhorst_pack((4, 4, 4))

    # Do check
    symm = Symmetry(id_a, cell_cv, pbc_c, symmorphic=False)
    symm.analyze(spos_ac)
    ibzk_kc, w_k = symm.reduce(bzk_kc)[:2]
    assert len(symm.op_scc) == 48
    assert len(w_k) == 10
    a = 3 / 32.
    b = 1 / 32.
    c = 6 / 32.
    assert np.all(w_k == [a, b, a, c, c, a, a, a, a, b])
    assert not symm.op_scc.sum(0).any()

    # Rotate unit cell and check again:
    cell_cv = a / sqrt(2) * np.array([(1, 0, 0),
                                      (0.5, sqrt(3) / 2, 0),
                                      (0.5, sqrt(3) / 6, sqrt(2.0 / 3))])
    symm = Symmetry(id_a, cell_cv, pbc_c, symmorphic=False)
    symm.analyze(spos_ac)
    ibzkb_kc, wb_k = symm.reduce(bzk_kc)[:2]
    assert len(symm.op_scc) == 48
    assert abs(w_k - wb_k).sum() < 1e-14
    assert abs(ibzk_kc - ibzkb_kc).sum() < 1e-14
    assert not symm.op_scc.sum(0).any()

    bzk_kc = monkhorst_pack((3, 3, 3))
    symm = Symmetry(id_a, cell_cv, pbc_c)
    symm.analyze(spos_ac)
    ibzk_kc, w_k = symm.reduce(bzk_kc)[:2]
    assert len(symm.op_scc) == 24
    assert len(w_k) == 4
    assert abs(w_k * 27 - (1, 12, 6, 8)).sum() < 1e-14
    assert not symm.op_scc.sum(0).any()

    # Rocksalt Ni2O2
    a = 7.92
    x = 2. * np.sqrt(1. / 3.)
    y = np.sqrt(1. / 8.)
    z1 = np.sqrt(1. / 24.)
    z2 = np.sqrt(1. / 6.)
    cell_cv = a * np.array([(x, y, -z1), (x, -y, -z1), (x, 0., z2)])
    spos_ac = np.array([[0., 0., 0.],
                        [1. / 2., 1. / 2., 1. / 2.],
                        [1. / 4., 1. / 4., 1. / 4.],
                        [3. / 4., 3. / 4., 3. / 4.]])
    id_a = [1, 2, 3, 3]
    pbc_c = np.array([1, 1, 1], bool)
    bzk_kc = monkhorst_pack((2, 2, 2))

    # Do check
    symm = Symmetry(id_a, cell_cv, pbc_c, symmorphic=False)
    symm.analyze(spos_ac)
    ibzk_kc, w_k = symm.reduce(bzk_kc)[:2]
    assert len(symm.op_scc) == 12
    assert len(w_k) == 2
    assert np.all(w_k == [3 / 4., 1 / 4.])

    # AlF3
    a = 2.465250000
    b = 1.423312751
    c = 4.148733333
    a1 = np.array([a, -1.0 * b, c])
    a2 = np.array([0.0, 2.0 * b, c])
    a3 = np.array([-1.0 * a, -1.0 * b, c])
    cell_cv = np.array([a1, a2, a3])
    id_a = [1, 1, 2, 2, 2, 2, 2, 2]
    spos_ac = np.array([(0.000000000, 0.000000000, 0.000000000),
                        (0.500000000, 0.500000000, 0.500000000),
                        (0.647758346, 0.852241654, 0.250000000),
                        (0.852241654, 0.250000000, 0.647758346),
                        (0.250000000, 0.647758346, 0.852241654),
                        (0.352241654, 0.147758346, 0.750000000),
                        (0.147758346, 0.750000000, 0.352241654),
                        (0.750000000, 0.352241654, 0.147758346)])
    pbc_c = np.array([1, 1, 1], bool)
    bzk_kc = monkhorst_pack((3, 3, 3))

    # Do check
    symm = Symmetry(id_a, cell_cv, pbc_c, symmorphic=False)
    symm.analyze(spos_ac)
    ibzk_kc, w_k = symm.reduce(bzk_kc)[:2]
    assert len(symm.op_scc) == 12
    assert len(w_k) == 6
    assert np.all(w_k == [1 / 27., 2 / 9., 2 / 9., 2 / 9., 2 / 9., 2 / 27.])
    assert not symm.op_scc.sum(0).any()



from __future__ import annotations
import warnings

from typing import Any, Sequence

import numpy as np
from gpaw.typing import DTypeLike

parameter_functions = {}

"""
background_charge
external
reuse_wfs_method
"""


def input_parameter(func):
    """Decorator for input-parameter normalization functions."""
    parameter_functions[func.__name__] = func
    return func


def update_dict(default: dict, value: dict | None) -> dict[str, Any]:
    """Create dict with defaults + updates.

    >>> update_dict({'a': 1, 'b': 'hello'}, {'a': 2})
    {'a': 2, 'b': 'hello'}
    >>> update_dict({'a': 1, 'b': 'hello'}, None)
    {'a': 1, 'b': 'hello'}
    >>> update_dict({'a': 1, 'b': 'hello'}, {'c': 2})
    Traceback (most recent call last):
    ValueError: Unknown key: 'c'. Must be one of a, b
    """
    dct = default.copy()
    if value is not None:
        if not value.keys() <= default.keys():
            key = (value.keys() - default.keys()).pop()
            raise ValueError(
                f'Unknown key: {key!r}. Must be one of {", ".join(default)}')
        dct.update(value)
    return dct


class InputParameters:
    basis: Any
    charge: float
    convergence: dict[str, Any]
    dtype: DTypeLike | None
    eigensolver: dict[str, Any]
    experimental: dict[str, Any]
    gpts: None | Sequence[int]
    h: float | None
    hund: bool
    kpts: dict[str, Any]
    magmoms: Any
    mode: dict[str, Any]
    nbands: None | int | str
    parallel: dict[str, Any]
    poissonsolver: dict[str, Any]
    setups: Any
    soc: bool
    spinpol: bool
    symmetry: dict[str, Any]
    xc: dict[str, Any]

    def __init__(self, params: dict[str, Any], warn: bool = True):
        self.keys = sorted(params)

        for key in params:
            if key not in parameter_functions:
                raise ValueError(
                    f'Unknown parameter {key!r}.  Must be one of: ' +
                    ', '.join(parameter_functions))
        for key, func in parameter_functions.items():
            if key in params:
                param = params[key]
                if hasattr(param, 'todict'):
                    param = param.todict()
                value = func(param)
            else:
                value = func()
            self.__dict__[key] = value

        if self.h is not None and self.gpts is not None:
            raise ValueError("""You can't use both "gpts" and "h"!""")

        if self.experimental is not None:
            if self.experimental.pop('niter_fixdensity', None) is not None:
                warnings.warn('Ignoring "niter_fixdensity".')
            if self.experimental.pop('reuse_wfs_method', None) is not None:
                warnings.warn('Ignoring "reuse_wfs_method".')
            if 'soc' in self.experimental:
                warnings.warn('Please use new "soc" parameter.',
                              DeprecatedParameterWarning)
                self.soc = self.experimental.pop('soc')
            if 'magmoms' in self.experimental:
                warnings.warn('Please use new "magmoms" parameter.',
                              DeprecatedParameterWarning)
                self.magmoms = self.experimental.pop('magmoms')
                self.keys.append('magmoms')
                self.keys.sort()
            assert not self.experimental
            self.keys.remove('experimental')
            self.__dict__.pop('experimental')

        if self.mode.get('name') is None:
            if warn:
                warnings.warn(
                    ('Finite-difference mode implicitly chosen; '
                     'it will be an error to not specify a mode '
                     'in the future'),
                    DeprecatedParameterWarning)
            self.mode = dict(self.mode, name='fd')
        force_complex_dtype = self.mode.pop('force_complex_dtype', None)
        if force_complex_dtype is not None:
            if warn:
                self.dtype = complex if force_complex_dtype else None
                warnings.warn(
                    'Please use '
                    f'GPAW(dtype={self.dtype}, '
                    '...)',
                    DeprecatedParameterWarning,
                    stacklevel=3)
            self.keys.append('dtype')
            self.keys.sort()

    def __repr__(self) -> str:
        p = ', '.join(f'{key}={value!r}'
                      for key, value in self.items())
        return f'InputParameters({p})'

    def items(self):
        for key in self.keys:
            yield key, getattr(self, key)


@input_parameter
def basis(value=None):
    """Atomic basis set."""
    return value or {}


@input_parameter
def charge(value=0.0):
    return value


@input_parameter
def convergence(value=None):
    """Accuracy of the self-consistency cycle."""
    return value or {}


@input_parameter
def dtype(value=None):
    return value


@input_parameter
def eigensolver(value=None) -> dict:
    """Eigensolver."""
    if isinstance(value, str):
        value = {'name': value}
    if value and value['name'] != 'dav':
        warnings.warn(f'{value["name"]} not implemented.  Using dav instead')
        return {'name': 'dav'}
    return value or {}


@input_parameter
def experimental(value=None):
    return value


@input_parameter
def gpts(value=None):
    """Number of grid points."""
    return value


@input_parameter
def h(value=None):
    """Grid spacing."""
    return value


@input_parameter
def hund(value=False):
    """Using Hund's rule for guessing initial magnetic moments."""
    return value


@input_parameter
def kpts(value=None) -> dict[str, Any]:
    """Brillouin-zone sampling."""
    if value is None:
        value = {'size': (1, 1, 1)}
    elif not isinstance(value, dict):
        array = np.array(value)
        if array.shape == (3,):
            value = {'size': array}
        else:
            value = {'kpts': array}
    return value


@input_parameter
def magmoms(value=None):
    return value


@input_parameter
def maxiter(value=333):
    """Maximum number of SCF-iterations."""
    return value


@input_parameter
def mixer(value=None):
    return value or {}


@input_parameter
def mode(value=None):
    if value is None:
        return {'name': value}
    if isinstance(value, str):
        return {'name': value}
    gc = value.pop('gammacentered', False)
    assert not gc
    return value


@input_parameter
def nbands(value: str | int | None = None) -> str | int | None:
    """Number of electronic bands."""
    return value


@input_parameter
def occupations(value=None):
    return value


@input_parameter
def parallel(value: dict[str, Any] | None = None) -> dict[str, Any]:
    dct = update_dict({'kpt': None,
                       'domain': None,
                       'band': None,
                       'order': 'kdb',
                       'stridebands': False,
                       'augment_grids': False,
                       'sl_auto': False,
                       'sl_default': None,
                       'sl_diagonalize': None,
                       'sl_inverse_cholesky': None,
                       'sl_lcao': None,
                       'sl_lrtddft': None,
                       'use_elpa': False,
                       'elpasolver': '2stage',
                       'buffer_size': None,
                       'gpu': False},
                      value)
    return dct


@input_parameter
def poissonsolver(value=None):
    """Poisson solver."""
    return value or {}


@input_parameter
def random(value=False):
    return value


@input_parameter
def setups(value='paw'):
    """PAW datasets or pseudopotentials."""
    return value if isinstance(value, dict) else {'default': value}


@input_parameter
def soc(value=False):
    return value


@input_parameter
def spinpol(value=False):
    return value


@input_parameter
def symmetry(value='undefined'):
    """Use of symmetry."""
    if value == 'undefined':
        value = {}
    elif value is None or value == 'off':
        value = {'point_group': False, 'time_reversal': False}
    return value


@input_parameter
def xc(value='LDA'):
    """Exchange-Correlation functional."""
    if isinstance(value, str):
        return {'name': value}
    return value


class DeprecatedParameterWarning(FutureWarning):
    """Warning class for when a parameter or its value is deprecated."""



import pytest

import numpy as np
from scipy.special import spherical_jn

from gpaw.ffbt import spherical_bessel


@pytest.mark.parametrize('l', range(7))
def test_spherical_bessel(l):
    x_g = np.arange(101, dtype=float)
    jl_g = spherical_bessel(l, x_g)
    assert jl_g == pytest.approx(spherical_jn(l, x_g))
